import time
import threading

from smbus import SMBus
import com_picockpit.co2sensor.constants as CONST
from picockpit_client.pmqtt import PCPMQTTClient as PCPMQTTClient


class ComPicockpitCO2Sensor:
    def __init__(self,
                 global_run_flag,
                 resource_manager,
                 hooks):
        self.global_run_flag = global_run_flag
        self.hooks = hooks
        self.topic_app_prefix = "com.picockpit/co2sensor/"
        self.st = threading.Thread(target=self.stats_thread, daemon=True)
        self.run_flag = threading.Event()
        self.on = False
        self.resource_manager = resource_manager
        # self.addr = CONST.CO2.I2CADDRESS
        self.error = False
        self.error_message = ""
        # I2C Bus
        self.sensor = None
        # run_flag controls the execution of the thread;
        # it is set in the on_connect_hook!
        # self.run_flag.set()
        self.co2_is_initialized = False
        self.st.start()

    def publish_error(self):
        msg = {
            "$error": self.error_message
        }
        self.publish(
            topic='$state',
            message=msg,
            retain=False,
            overwrite=True)
        # force a delay of 4 seconds to slow down messages 
        time.sleep(4)

    def sigterm_hook(self):
        # allow the internal thread which might be blocked to join.
        self.run_flag.set()

    def subscribe(self, topic, message_callback_function):
        self.hooks["subscribe"](
            topic=self.topic_app_prefix + topic,
            message_callback_function=message_callback_function)

    def unsubscribe(self, topic):
        self.hooks["unsubscribe"](
            topic=self.topic_app_prefix + topic)

    def publish(self, topic, message, retain=False, overwrite=True):
        self.hooks["publish"](
            topic_app_prefix=self.topic_app_prefix,
            topic_proper=topic,
            message=message,
            retain=retain,
            overwrite=overwrite)

    def stats_thread(self):
        ld = threading.local()
        while self.global_run_flag.is_set():
            # block until run_flag is true
            # (we have just reconnected or have received a keep alive)
            # this run_flag will also be unblocked on the sigterm_hook being called
            # allow the tread to terminate by falling out at the bottom
            self.run_flag.wait()
            # time.time() returns number of seconds since the epoch
            # using int to cast to an integer without fractions of a second
            ld.ts_last_keep_alive = int(time.time())
            ld.ts_now = int(time.time())
            # three minutes = 180 seconds
            while (ld.ts_now - ld.ts_last_keep_alive < 180) and self.global_run_flag.is_set():
                # the flag will be set to true by the initial setup
                # and by periodic pings from the picockpit frontend
                # whenever it is set to true, we reset the last_keep_alive
                # and clear the flag (set it to false)
                # this way, if no further keep alives arrive, the system
                # will drop into blocking / waiting mode at some point
                if self.run_flag.is_set():
                    ld.ts_last_keep_alive = int(time.time())
                    self.run_flag.clear()
                # sys.stdout.write("✪")
                # sys.stdout.flush()
                if self.on:
                    if self.co2_is_initialized:
                        if self.resource_manager.i2c.i2c_active:
                            self.publish(
                                topic='$state',
                                message=self.get_stats(),
                                retain=False,
                                overwrite=True)
                        else:
                            self.error_message = "I2C is not active - please enable it"
                            self.publish_error()
                    else:
                        if self.resource_manager.i2c.i2c_active:
                            try:
                                self.sensor = SMBus(1)
                                self.co2_is_initialized = True
                            except Exception:
                                self.error_message = "com.picockpit/co2sensor: I2C problem. Please try to disable and reenable the sensor."
                                self.publish_error()                            
                        else:
                            self.error_message = "I2C is or was not active - please try disabling and reenabling"
                            self.publish_error()                        
                time.sleep(1)
                ld.ts_now = int(time.time())

    # returns a Dict with the values in it, which can be published directly
    # TODO: rewrite for co2688 Airmeter
    def get_stats(self):
        stats = {}
        # Todo: only get_iaq() and appropriate entries if self.iaq == True
        try:
            read_data = self.sensor.read_i2c_block_data(CONST.EE895ADDRESS, CONST.I2CREGISTER)
            co2 = read_data[0].to_bytes(1, 'big') + read_data[1].to_bytes(1, 'big')
            temperature = read_data[2].to_bytes(1, 'big') + read_data[3].to_bytes(1, 'big')
            pressure = read_data[6].to_bytes(1, 'big') + read_data[7].to_bytes(1, 'big')
            stats["t"] = int.from_bytes(temperature, "big") / 100
            stats["p"] = int.from_bytes(pressure, "big") / 10
            stats["c"] = int.from_bytes(co2, "big")
        except Exception:
            stats["$error"] = "com.picockpit/co2sensor: Could not read from the sensor. Is it attached?"
            # force some additional sleep to avoid flooding PiCockpit's MQTT server
            time.sleep(4)
        return stats

    def timer_hook(self, timer_spacing):
        pass

    def on_set_msg(self, client, userdata, message):
        topic, payload = PCPMQTTClient.parse_message(message)
        if "on" in payload:
            if payload["on"] is True:
                if not self.run_flag.is_set():
                    self.on = True
                    # gratutious initial supply with data
                    self.run_flag.set()
                    if not self.co2_is_initialized:
                        if self.resource_manager.i2c.i2c_active:
                            try:
                                self.sensor = SMBus(1)
                                self.co2_is_initialized = True
                            except Exception:
                                self.error_message = "com.picockpit/co2sensor: I2C problem. Please try to disable and reenable the sensor."
                                self.publish_error()                            
                        else:
                            self.error_message = "com.picockpit/co2sensor: I2C seems to be disabled. Please try to disable and reenable the sensor."
                            self.publish_error()
            else:
                self.on = False
                if self.co2_is_initialized:
                    self.sensor.close()
                    self.co2_is_initialized = False
                self.run_flag.clear()
        else:
            self.on = False
            self.run_flag.clear()

    def on_control_msg(self, client, userdata, message):
        # this resets the thread's countdown.
        # not interested in the message contents
        self.run_flag.set()

    def reload_hook(self):
        # avoid double subscription (possibly unnecessary here)
        self.unsubscribe(topic="control")
        self.unsubscribe(topic="set")
        # this possibly should ensure that the client is connected ..
        # or better: ensure upstream that no errors are generated by this
        self.on_connect_hook()

    def publish_meta(self):
        meta = {
            "t": {
                "$description": "air temperature",
                "$unit": "°C",
                "$min": -40,
                "$max": 60
            },
            "p": {
                "$description": "air pressure",
                "$unit": "hPa",
                "$min": 700,
                "$max": 1100
            },
            "c": {
                "$description": "CO2",
                "$unit": "ppm",
                "$min": 0,
                "$max": 5000
            }
        }
        self.publish(
            topic='$meta',
            message=meta,
            retain=True,
            overwrite=True)

    def on_connect_hook(self):
        self.subscribe(
            topic="control",
            message_callback_function=self.on_control_msg)
        self.subscribe(
            topic="set",
            message_callback_function=self.on_set_msg)
        self.publish_meta()
