# import sys
import dbus
import time
import netifaces
from picockpit_client._version import version

from picockpit_client.pmqtt import PCPMQTTClient as PCPMQTTClient


class ComPicockpitCore:
    def __init__(self,
                 global_run_flag,
                 resource_manager,
                 hooks,
                 getpublicip,
                 terminate):
        self.startuptime = int(time.time())
        self.connection_counter = 0
        self.reconnect = False
        self.global_run_flag = global_run_flag
        self.publish = hooks['publish']
        self.topic_app_prefix = 'com.picockpit/core/'
        self.subscribe = hooks['subscribe']
        self.unsubscribe = hooks['unsubscribe']
        self.terminate = terminate
        # access to resource manager class
        self.resource_manager = resource_manager
        self.resources_subscribed = False
        self.getpublicip = getpublicip

    def getip_info(self):
        interfaces = netifaces.interfaces()
        rethash = {}
        # not doing gateways for the moment.
        for interface in interfaces:
            rethash[interface] = netifaces.ifaddresses(interface)
            # rewriting the keys to fixed references, instead of numbers
            if netifaces.AF_LINK in rethash[interface]:
                rethash[interface]['AF_LINK'] = rethash[interface][netifaces.AF_LINK]
                del rethash[interface][netifaces.AF_LINK]
            if netifaces.AF_INET in rethash[interface]:
                rethash[interface]['AF_INET'] = rethash[interface][netifaces.AF_INET]
                del rethash[interface][netifaces.AF_INET]
            if netifaces.AF_INET6 in rethash[interface]:
                rethash[interface]['AF_INET6'] = rethash[interface][netifaces.AF_INET6]
                del rethash[interface][netifaces.AF_INET6]
        return rethash

    def timer_hook(self, timer_spacing):
        pass

    def on_gpio_mqtt_msg(self, client, userdata, message):
        PCPMQTTClient.parse_message(message)

    def on_delete_mqtt_msg(self, client, userdata, message):
        print("Disabling picockpit-client daemon service ...")
        system_bus = dbus.SystemBus()
        systemd1 = system_bus.get_object('org.freedesktop.systemd1', '/org/freedesktop/systemd1')
        manager = dbus.Interface(systemd1, 'org.freedesktop.systemd1.Manager')
        manager.DisableUnitFiles(['picockpit-client.service'], False)
        print("Goodbye, we hope to see you back soon :-)")
        manager.Reload()
        self.terminate()

    def on_setup_apps_mqtt_msg(self, client, userdata, message):
        topic, payload = PCPMQTTClient.parse_message(message)
        # we need to bring up a structure which matches apps with UUIDs
        if 'apps' in payload:
            self.resource_manager.apps.update(payload['apps'])
            if not self.resources_subscribed:
                self.subscribe_resources()

    def subscribe_resources(self):
        self.subscribe(
            topic="com.picockpit/core/resources/set/gpio",
            message_callback_function=self.on_resources_set_gpio_mqtt_msg)
        self.subscribe(
            topic="com.picockpit/core/resources/set/i2c",
            message_callback_function=self.on_resources_set_i2c_mqtt_msg)
        self.resources_subscribed = True

    def on_resources_set_i2c_mqtt_msg(self, client, userdata, message):
        topic, payload = PCPMQTTClient.parse_message(message)
        # let resource manager I2C take care of the remaining stuff
        if 'i2cActive' in payload:
            self.resource_manager.i2c.set_state(payload['i2cActive'])

    def on_resources_set_gpio_mqtt_msg(self, client, userdata, message):
        topic, payload = PCPMQTTClient.parse_message(message)
        # let resource manager GPIO take care of the remaining stuff
        if 'data' in payload:
            self.resource_manager.gpio.update(payload['data'])
            # also update the I2C state
            # True -> only publish an update if there are changes.
            self.resource_manager.i2c.publish_state(True)

    def on_generic_message_callback(self, client, userdata, message):
        pass

    def reload_hook(self):
        self.resources_subscribed = False
        self.unsubscribe(topic="com.picockpit/core/resources/set/gpio")
        self.unsubscribe(topic="com.picockpit/core/setup/set/apps")
        # self.unsubscribe(topic="com.picockpit/core/$overlay/+/$set/gpio")
        self.on_connect_hook()

    def publish_client_state(self, run=True):
        # version is REQUIRED!
        publish_client = {
            "version": version,
            "startuptime": self.startuptime,
            "connectionCounter": self.connection_counter,
            "run": run
        }

        self.publish(
            topic_app_prefix=self.topic_app_prefix,
            topic_proper='$state/client',
            message=publish_client,
            retain=True,
            overwrite=True)

    def set_online(self):
        # NB this is an exception, of us publishing on a main path,
        # not on a subpath - this indicates the whole system is up
        self.publish(
            topic_app_prefix='',
            topic_proper='$state',
            message={"onlineState": "online"},
            qos=2,
            retain=True,
            overwrite=True)

    def on_disconnect_hook(self):
        self.resources_subscribed = False

    def on_connect_hook(self):
        self.set_online()
        self.connection_counter += 1
        self.publish_client_state()
        # subscribe to GPIO in core
        # the $state/gpio path is one we publish to. no need to subscribe to it
        # and to GPIO overlays
        # TODO: remove old setup
        # self.subscribe(
        #     topic="com.picockpit/core/$overlay/+/$set/gpio",
        #     message_callback_function=self.on_gpio_overlay_mqtt_msg)

        self.subscribe(
            topic="com.picockpit/core/setup/set/apps",
            message_callback_function=self.on_setup_apps_mqtt_msg)

        self.subscribe(
            topic="delete",
            message_callback_function=self.on_delete_mqtt_msg)

        publicip = ""
        apianswer = self.getpublicip()
        if apianswer["error"]:
            print("ERROR: unable to obtain public ip")
            print("ANSWER BODY:")
            print(apianswer["body"])
        else:
            publicip = apianswer["object"]["publicip"]

        pinfo = self.getip_info()
        privateips = []
        # TODO : decide whether we want accurate info or just rough info!
        for interface, content in pinfo.items():
            if not interface == "lo":
                if "AF_INET" in content:
                    privateips.append({
                        "name": interface,
                        "ip": content["AF_INET"][0]["addr"]})

        publish_set = {
            "public": publicip,
            "private": privateips
        }
        # TODO: do we REALLY want retain here? (on IP addresses)
        self.publish(
            topic_app_prefix=self.topic_app_prefix,
            topic_proper='$state/network',
            message=publish_set,
            retain=True)

        # also update the I2C state
        self.resource_manager.i2c.publish_state()

    def announce_shutdown(self):
        self.publish_client_state(run=False)

    def teardown_pin(self, bcm_id):
        # no-op, to satisfy the current system for tearing down I2C pins.
        pass
