import com_picockpit.constants as CONST
import com_picockpit.gpio.gpio_local_constants as GPIO_LOCAL_CONST
from picockpit_client.pmqtt import PCPMQTTClient as PCPMQTTClient

try:
    import RPi.GPIO as GPIO
except RuntimeError:
    print("Error importing RPi.GPIO!  This is probably because you need superuser privileges.  You can achieve this by using 'sudo' to run your script")


class GpioPinIO:
    def __init__(self, bcm_id, hooks):
        self.bcm_id = bcm_id
        self.publish = hooks['publish']
        self.subscribe = hooks['subscribe']
        self.unsubscribe = hooks['unsubscribe']
        self.topic_app_prefix = 'com.picockpit/gpio/'
        self.publish_topic = "$state/" + str(self.bcm_id)
        self.subscribe_topic = "com.picockpit/gpio/set/" + str(self.bcm_id)
        # 11.9.2020; this topic should be subscribed to when the pin is activated. 
        self.pin_setup_topic = "com.picockpit/gpio/set/" + str(self.bcm_id) + "/setup"
        self.pin_setup_subscribed = False
        self.pin_setup_resubscribe_necessary = False
        self.set_defaults()

    def set_defaults(self):
        self.mode = CONST.GpioCapabilities.UNMANAGED
        # pull is initially not defined and needs to be set.
        self.pull = CONST.GpioPull.UNDEFINED
        self.output_mode = CONST.GpioOutputMode.DEFAULT
        self.pwm_handle = None
        self.frequency = 1
        self.duty_cycle = 100
        self.debounce_time = CONST.GpioDefaults.DEBOUNCE

    def set_pull(self, pull):
        if (pull != self.pull) and self.mode == CONST.GpioCapabilities.INPUT:
            pull_gpio = GPIO.PUD_OFF
            if pull == CONST.GpioPull.UP:
                pull_gpio = GPIO.PUD_UP
            elif pull == CONST.GpioPull.DOWN:
                pull_gpio = GPIO.PUD_DOWN
            GPIO.setup(self.bcm_id, GPIO.IN, pull_up_down=pull_gpio)
            self.pull = pull

    def process_input_settings(self, data):
        if 'debounce' in data:
            self.update_debounce_time(data['debounce'])
        fixed_pull_up = False
        if 'fixedPullUp' in data:
            # TODO: type check for bool
            fixed_pull_up = data['fixedPullUp']
        if not fixed_pull_up:
            use_default_pull = False
            if 'useDefaultPull' in data:
                # TODO: type check for bool
                use_default_pull = data['useDefaultPull']
            if use_default_pull and 'defaultPull' in data:
                self.set_pull(data['defaultPull'])
            elif 'pull' in data:
                self.set_pull(data['pull'])
            # free update after changing anything
        self.update_gpio_state()

    def process_output_settings(self, data):
        if 'target' in data:
            if data['target'] == GPIO_LOCAL_CONST.TargetDef.OUTPUT:
                # set pin as regular output pin
                self.output_mode = CONST.GpioOutputMode.DEFAULT
                self.update_gpio_state()
            elif data['target'] == GPIO_LOCAL_CONST.TargetDef.SOFTPWM:
                # pin as SoftPWM pin
                self.output_mode = CONST.GpioOutputMode.SOFTPWM

    def on_pin_setup_mqtt_msg(self, client, userdata, message):
        topic, payload = PCPMQTTClient.parse_message(message)
        # process
        if 'data' in payload:
            data = payload['data']
            # for INPUT the target does not matter - there is just one
            if self.mode == CONST.GpioCapabilities.INPUT:
                self.process_input_settings(data)
            elif self.mode == CONST.GpioCapabilities.OUTPUT:
                self.process_output_settings(data)

    def on_connect_hook(self):
        if self.pin_setup_subscribed and self.pin_setup_resubscribe_necessary:
            self.subscribe(
                topic=self.pin_setup_topic,
                message_callback_function=self.on_pin_setup_mqtt_msg)
            if self.mode == CONST.GpioCapabilities.OUTPUT:
                self.subscribe(self.subscribe_topic,
                               self.mqtt_output_control_callback)
        self.pin_setup_resubscribe_necessary = False

    def on_disconnect_hook(self):
        if self.pin_setup_subscribed:
            self.pin_setup_resubscribe_necessary = True

    def setup(self, mode):
        # set up sane defaults
        self.set_defaults()
        if mode != CONST.GpioCapabilities.UNMANAGED and not self.pin_setup_subscribed:
            self.pin_setup_subscribed = True
            self.subscribe(
                topic=self.pin_setup_topic,
                message_callback_function=self.on_pin_setup_mqtt_msg)

        # initial setup until we receive a different control message
        if mode == CONST.GpioCapabilities.INPUT:
            self.mode = CONST.GpioCapabilities.INPUT
            self.update_gpio_state()
            GPIO.add_event_detect(self.bcm_id,
                                  GPIO.BOTH,
                                  callback=self.gpio_edge_callback,
                                  bouncetime=self.debounce_time)
        elif mode == CONST.GpioCapabilities.OUTPUT:
            self.mode = CONST.GpioCapabilities.OUTPUT
            # consider: should we drive the pin low initially? (possibly better to keep states?)
            self.subscribe(self.subscribe_topic,
                           self.mqtt_output_control_callback)

    def mqtt_output_control_callback(self, client, userdata, message):
        topic, payload = PCPMQTTClient.parse_message(message)
        if self.mode == CONST.GpioCapabilities.OUTPUT:
            if self.output_mode == CONST.GpioOutputMode.DEFAULT:
                if "state" in payload:
                    if payload["state"] == CONST.GpioState.HIGH:
                        GPIO.output(self.bcm_id, GPIO.HIGH)
                    elif payload["state"] == CONST.GpioState.LOW:
                        GPIO.output(self.bcm_id, GPIO.LOW)
                    self.update_gpio_state()
            elif self.output_mode == CONST.GpioOutputMode.SOFTPWM:
                if "freq" in payload and "dc" in payload:
                    new_frequency = None
                    new_duty_cycle = None
                    if type(payload["freq"]) == int or type(payload["freq"]) == float:
                        new_frequency = payload["freq"]
                    if type(payload["dc"]) == int or type(payload["dc"]) == float:
                        new_duty_cycle = payload["dc"]
                    if self.pwm_handle is None:
                        if new_frequency is not None:
                            self.frequency = new_frequency
                        if new_duty_cycle is not None:
                            self.duty_cycle = new_duty_cycle
                        # GPIO.output(self.bcm_id, GPIO.HIGH)
                        self.pwm_handle = GPIO.PWM(self.bcm_id, self.frequency)
                        self.pwm_handle.start(self.duty_cycle)
                    else:
                        if new_frequency is not None:
                            if new_frequency != self.frequency:
                                self.frequency = new_frequency
                                self.pwm_handle.ChangeFrequency(self.frequency)
                        if new_duty_cycle is not None:
                            if new_duty_cycle != self.duty_cycle:
                                self.duty_cycle = new_duty_cycle
                                self.pwm_handle.ChangeDutyCycle(self.duty_cycle)

    def gpio_edge_callback(self, channel):
        # only report if we are still in the right mode
        if self.mode == CONST.GpioCapabilities.INPUT:
            # true or false
            state = CONST.GpioState.LOW
            if GPIO.input(channel):
                state = CONST.GpioState.HIGH
            data = {
                "state": state
            }
            # Todo: rethink decision for retain ...
            self.publish(
                topic_app_prefix=self.topic_app_prefix,
                topic_proper=self.publish_topic,
                message=data,
                retain=True,
                overwrite=True)

    def set_state_unmanaged(self):
        data = {
            "state": CONST.GpioState.UNMANAGED
        }
        # self.publish(self.publish_topic, data, retain=True)
        self.publish(
            topic_app_prefix=self.topic_app_prefix,
            topic_proper=self.publish_topic,
            message=data,
            retain=True,
            overwrite=True)

    def tear_down(self):
        if self.mode == CONST.GpioCapabilities.INPUT:
            GPIO.remove_event_detect(self.bcm_id)
            self.set_state_unmanaged()
        elif self.mode == CONST.GpioCapabilities.OUTPUT:
            self.unsubscribe(self.subscribe_topic)
            if self.pwm_handle is not None:
                self.pwm_handle.stop()
                self.pwm_handle = None
            else:
                self.set_state_unmanaged()

        self.mode = CONST.GpioCapabilities.UNMANAGED
        self.pin_setup_subscribed = False
        self.pin_setup_resubscribe_necessary = False
        self.unsubscribe(topic=self.pin_setup_topic)

    def update_gpio_state(self):
        state = CONST.GpioState.LOW
        if GPIO.input(self.bcm_id):
            state = CONST.GpioState.HIGH
        data = {
            "state": state
        }
        # Todo: rethink decision for retain ...
        self.publish(
            topic_app_prefix=self.topic_app_prefix,
            topic_proper=self.publish_topic,
            message=data,
            retain=True,
            overwrite=True)

    def update_debounce_time(self, debounce_time):
        if (isinstance(debounce_time, int)):
            if (debounce_time < 1):
                debounce_time = 1
            # only update if time is different
            if (self.debounce_time != debounce_time):
                self.debounce_time = debounce_time
                GPIO.remove_event_detect(self.bcm_id)
                GPIO.add_event_detect(self.bcm_id,
                                      GPIO.BOTH,
                                      callback=self.gpio_edge_callback,
                                      bouncetime=self.debounce_time)
