import os
import re
import pwd
import copy
import glob
import json
import time
import errno
import psutil
import signal
# import datetime
import threading
import subprocess

from string import Template
from queue import PriorityQueue
from picockpit_client.pmqtt import PCPMQTTClient as PCPMQTTClient

CONFIGPATH = '/etc/picockpit-client/apps/com.picockpit/picontrol/modules/'


class WorkerThread:
    def __init__(self, module_handle, command_handle, config, publish_hook):
        # this will be managed by publish_state which gets called by the Thread!
        self.current_stage = 0
        self.module_handle = module_handle
        self.command_handle = command_handle
        self.publish = publish_hook
        # reaction to commands will be disabled on shutdown.
        self.listen = True
        self.notermination = False
        self.config = {}
        self.terminate_flag = False
        # publishing the state in init is too early! there is no MQTT object yet!

        # the flag is initially false
        # run_flag will be set if the command is executing
        self.run_flag = threading.Event()
        # inverse run flag
        self.i_run_flag = threading.Event()
        self.i_run_flag.set()
        self.wt = threading.Thread(target=self.thread_run, daemon=True)
        # stop_flag will be set if the command should stop
        self.stop_flag = threading.Event()
        self.wt.start()
        self.refresh(config=config)

    def publish_state(self,
                      stage=None,
                      have_run_flag=False,
                      stopped_flag=False,
                      killed_flag=False,
                      terminating_flag=False,
                      failed_flag=False):
        if stage is None:
            stage = self.current_stage
        else:
            self.current_stage = stage
        msg = {
            "run": self.run_flag.is_set()
        }
        if self.run_flag.is_set():
            msg["stage"] = stage
        if have_run_flag:
            msg["finished"] = True
        if terminating_flag:
            msg["terminating"] = True
        if stopped_flag:
            msg["stopped"] = True
        if killed_flag:
            msg["killed"] = True
        if failed_flag:
            msg["failed"] = True

        topic = "$state/" + self.module_handle + "/" + self.command_handle
        self.publish(
            topic=topic,
            message=msg,
            retain=True,
            overwrite=True)

    def stop(self):
        self.listen = False
        # if the command was not executing, no need to announce that it is being terminated.
        if self.run_flag.is_set():
            self.publish_state(terminating_flag=True)
        self.stop_flag.set()

    # this is more of a rough patch and should be rewritten / reimagined
    def terminate(self):
        self.terminate_flag = True
        # unblock the threads so they can join
        self.run_flag.set()

    def wait_shutdown(self):
        # note: this will block on commands which have "0" termination timeout
        # TODO: come up with an alternative timeout / solution?
        self.i_run_flag.wait()

    def refresh(self, config):
        # TODO: refresh will have to reread the config, update timeout etc!
        # TODO: refresh has to re-set the command definitions
        # this gets the new config, and should check whether it has changed and we need to terminate the process if it runs

        # wait with the refresh until the thread is not executing anymore.
        # this prevents all kinds of problems with accessing data structures simultaneously
        # i_run_flag is an inverted run_flag; to be unset where run_flag is set and vice versa
        self.i_run_flag.wait()
        self.config = config
        # notermination influences whether we try to shut down the command in any way
        if "notermination" in self.config:
            if isinstance(self.config["notermination"], bool):
                self.notermination = self.config["notermination"]
        self.commands = []
        # command_handle = self.command_handle
        # handle nested and normal commands, to be represented as the same thing
        if "command" in self.config:
            if isinstance(self.config["command"], list):
                if isinstance(self.config["command"][0], list):
                    self.commands = self.config["command"]
                else:
                    self.commands = [self.config["command"]]
        self.fresh_flag = True

    def set_stale(self):
        self.fresh_flag = False

    def process_stale_state(self):
        if not self.fresh_flag:
            # TODO: Problem is that refresh waits for the i_run_flag; in the meantime this could get
            # executed - and therefore stop the thread, as refresh is not able to set the fresh_flag!
            # this needs to be thought through more carefully.
            # if self.run_flag.is_set():
            #     self.publish_state(terminating_flag=True)
            #     self.stop_flag.set()
            # TODO: make sure the command stops, even if kill timeout is 0
            # print("here we would terminate the command if it executes currently, etc.")
            pass

    def process_message(self, topic, payload):
        if self.listen:
            # message can be run: true or run: false
            if "run" in payload:
                # silently discard double messages
                if payload["run"] is True and not self.run_flag.is_set():
                    # clearing stop flag if it has been set before
                    self.stop_flag.clear()
                    self.i_run_flag.clear()
                    self.run_flag.set()
                else:
                    if not self.notermination:
                        self.publish_state(terminating_flag=True)
                        self.stop_flag.set()

    def thread_run(self):
        def sanitize(input_string):
            #https://maxharp3r.wordpress.com/2008/05/15/pythons-minidom-xml-and-illegal-unicode-characters/
            RE_XML_ILLEGAL = u'([\u0000-\u0008\u000b-\u000c\u000e-\u001f\ufffe-\uffff])' + \
                             u'|' + \
                             u'([%s-%s][^%s-%s])|([^%s-%s][%s-%s])|([%s-%s]$)|(^[%s-%s])' % \
                             (chr(0xd800), chr(0xdbff), chr(0xdc00), chr(0xdfff),
                              chr(0xd800), chr(0xdbff), chr(0xdc00), chr(0xdfff),
                              chr(0xd800), chr(0xdbff), chr(0xdc00), chr(0xdfff))
            return re.sub(RE_XML_ILLEGAL, "", input_string)

        def demote(user, uid, gid):
            def demote_function():
                # initgroups must be run before we lose the privilege to set it!
                os.initgroups(user, gid)
                os.setgid(gid)
                # this must be run last
                os.setuid(uid)
            return demote_function

        def terminate_recursively(proc_pid, kill=False):
            # return false if the process does not exist anymore
            if not psutil.pid_exists(proc_pid):
                # print("picontrol :: terminate_recursively :: pid does not exist" + str(proc_pid) + " ... returning False")
                return False
            # https://stackoverflow.com/questions/4789837/how-to-terminate-a-python-subprocess-launched-with-shell-true
            try:
                parent = psutil.Process(proc_pid)
                # process = psutil.Process(proc_pid)
                children = parent.children(recursive=True)
                children.append(parent)
                sig = signal.SIGTERM
                if kill:
                    sig = signal.SIGKILL
                for proc in children:
                    proc.send_signal(sig)
            except Exception as e:
                print("picontrol :: terminate_recursively failed, " + str(e))
                # TODO: possibly here the psutil convenience function could be added, and used to collect shutdown info 
                # from all the processes
                # BUT: this would require a rewrite of the timing / timeout routine.
                    # if kill:
                    #     proc.kill()
                    # else:
                    #     proc.terminate()
                # if kill:
                #     process.kill()
                # else:
                #     process.terminate()
            return True

        def run_cmd(cmd, cmd_def):
            return_code = 0
            return_error = ""
            return_result = ""
            timeout = None
            terminatetimeout = 10
            user = "pi"
            if "timeout" in cmd_def:
                if isinstance(cmd_def["timeout"], int):
                    if cmd_def["timeout"] > 0:
                        timeout = cmd_def["timeout"]
            if "terminatetimeout" in cmd_def:
                if isinstance(cmd_def["terminatetimeout"], int):
                    # this must be allowed to become 0!
                    timeout = cmd_def["terminatetimeout"]
            if "user" in cmd_def:
                if isinstance(cmd_def["user"], str):
                    user = cmd_def["user"]
            try:
                uid = pwd.getpwnam(user).pw_uid
                gid = pwd.getpwnam(user).pw_gid
                proc = psutil.Popen(
                    cmd,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    preexec_fn=demote(user, uid, gid)
                )
                # TODO: handle timeout differently
                    # timeout=timeout)                
                process = psutil.Process(proc.pid)
                start_time = process.create_time()
            except FileNotFoundError:
                return_code = errno.ENOENT
                return_error = "Command not found on your system " + " ".join(cmd)
            # TODO: will this actually ever be called?
            except subprocess.TimeoutExpired:
                return_code = errno.ETIME
                return_error = "Command timeout expired (" + str(cmd_def["timeout"]) + " sec)"
            except Exception as e:
                if hasattr(e, 'errno'):
                    return_code = e.errno
                else:
                    return_code = 1
                if hasattr(e, 'message'):
                    return_error = str(e.message)
                else:
                    return_error = str(e)
            else:
                # process can execute normally, no exceptions at startup
                process_running = True
                terminating = False
                # stopping = False
                killing = False
                termination_time = None
                # if "startuptime" in cmd_def:
                #     if isinstance(cmd_def["startuptime"], int):
                #         if cmd_def["startuptime"] > 0:
                #             print("picontrol :: startup time detected, starting to sleep")
                #             time.sleep(cmd_def["startuptime"])
                #             print("picontrol :: startup time finished.")
                # return_error = sanitize(proc.stderr.read().decode('UTF-8').rstrip("\r\n"))
                # print("picontrol :: early return_error")
                # print(return_error)
                # return_result = sanitize(proc.stdout.read().decode('UTF-8').rstrip("\r\n"))
                # print("picontrol :: early return_result")
                # print(return_result)
                while process_running:
                    if proc.poll() is not None:
                        process_running = False
                    if self.stop_flag.is_set() and not terminating and not self.notermination:
                        # stopping = True
                        terminating = terminate_recursively(proc.pid)
                        # if terminate_recursively returns false, there was nothing to terminate -> we can exit the loop
                        if not terminating:
                            process_running = False
                        else:
                            termination_time = time.time()
                        # avoid double termination attempt
                        # terminating = True
                    # do steps for active termination after that.
                    # we make use of the fact that 0 is evaluated to false on Python (thus deactivating termination)
                    # if self.stop_flag.is_set() and terminating and terminatetimeout:
                    # this is used for the regular timeout as well, if command does not quit on time
                    if terminating and terminatetimeout:
                        if (time.time() - termination_time) > terminatetimeout:
                            # killing = True
                            killing = terminate_recursively(proc.pid, kill=True)
                            # if terminate_recursively returns false, there was nothing to kill -> we can exit the loop
                            if not killing:
                                process_running = False                            
                    if process_running:
                        # one tenth of a second of resolution for faster feedback
                        # only need to sleep if the process is running.
                        time.sleep(0.1)
                        if timeout and not self.notermination:
                            if (time.time() - start_time) > timeout:
                                # stopping = True
                                terminating = terminate_recursively(proc.pid)
                                # if terminate_recursively returns false, there was nothing to terminate -> we can exit the loop
                                if not terminating:
                                    process_running = False
                                else:
                                    termination_time = time.time()
                # after process_running
                stop_flag = False
                kill_flag = False
                if killing:
                    kill_flag = True
                # elif stopping:
                elif terminating:
                    stop_flag = True
                return_code = proc.returncode
                # if not self.stop_flag.is_set():
                return_error = sanitize(proc.stderr.read().decode('UTF-8').rstrip("\r\n"))
                if not return_error:
                    if kill_flag:
                        if self.stop_flag.is_set():
                            return_error = "Killed process on user request (SIGKILL)."
                        else:
                            return_error = "Killed process on timeout (SIGKILL)."
                    elif stop_flag:
                        if self.stop_flag.is_set():
                            return_error = "Terminated process on user request (SIGTERM)."
                        else:
                            return_error = "Terminated process on timeout (SIGTERM)."                            
                return_result = sanitize(proc.stdout.read().decode('UTF-8').rstrip("\r\n"))
            return (return_code, return_error, return_result, stop_flag, kill_flag)

        ld = threading.local()
        ld.run = True
        while ld.run:
            self.run_flag.wait()
            # the terminate_flag will be set on shutdown to unblock.
            if self.terminate_flag:
                ld.run = False
            else:
                ld.stopped = False
                ld.killed = False
                ld.stage = -1
                ld.have_run_flag = False
                ld.failed_flag = False
                # if run_flag is set, we start command execution
                # first, publish state (that we started execution)
                cmd_def = self.config
                for cmd in self.commands:
                    if self.stop_flag.is_set():
                        break
                    ld.stage += 1
                    self.publish_state(stage=ld.stage)
                    return_code, return_error, return_result, stop_flag, kill_flag = run_cmd(
                        cmd=cmd,
                        cmd_def=cmd_def
                    )
                    if return_code != 0:
                        ld.failed_flag = True
                    ld.stopped = stop_flag
                    ld.killed = kill_flag
                    msg = {
                        "data": {
                            "returncode": return_code,
                            "error": return_error,
                            "output": return_result
                        }
                    }
                    ld.have_run_flag = True
                    # publish_state is called before execution
                    # here we publish the result of the execution for this stage
                    self.publish(
                        topic="$state/" + self.module_handle + "/" + self.command_handle + "/" + str(ld.stage),
                        message=msg,
                        overwrite=True
                    )
                # TODO: publish a message if the command is EMPTY -> so nothing could be executed
                self.i_run_flag.set()
                self.run_flag.clear()
                # this is publishing that the operation has completed.
                self.publish_state(have_run_flag=ld.have_run_flag,
                                   killed_flag=ld.killed,
                                   stopped_flag=ld.stopped,
                                   failed_flag=ld.failed_flag)


class ComPicockpitPiControl:
    """ This class runs commands on the Pi, and returns their results to picockpit.com
        It executes commands as separate threads, deploying them from the message queue
        the picockpit frontend should disable a button when a command is being run currently
        otherwise we let the system take care of concurrencies (e.g. similar to multiple SSH sessions being open) 
    """
    def __init__(self, global_run_flag, hooks):
        self.global_run_flag = global_run_flag
        self.hooks = hooks
        # self.master_topic_path = "com.picockpit/picontrol/"
        self.topic_app_prefix = "com.picockpit/picontrol/"
        self.topic_regex = re.compile('^pi/[^/]+/com.picockpit/picontrol/control/([^/]+)/?(.*)')
        # this is a dict of dicts: modules -> commands -> definitions
        self.worker_threads = {}
        self.reload_modules()

    def subscribe(self, topic, message_callback_function):
        self.hooks["subscribe"](
            topic=self.topic_app_prefix + topic,
            message_callback_function=message_callback_function)

    def unsubscribe(self, topic):
        self.hooks["unsubscribe"](
            topic=self.topic_app_prefix + topic)

    def publish(self, topic, message, retain=False, overwrite=False):
        self.hooks["publish"](
            topic_app_prefix=self.topic_app_prefix,
            topic_proper=topic,
            message=message,
            retain=retain,
            overwrite=overwrite)

    def timer_hook(self, timer_spacing):
        pass

    def sigterm_hook(self):
        # first stop all commands
        for module_name in self.worker_threads:
            for command_handle in self.worker_threads[module_name]:
                self.worker_threads[module_name][command_handle].stop()
                self.worker_threads[module_name][command_handle].terminate()
        # wait until all commands have shut down.
        #  TODO 2.8.2020: this seems to block teardown of the app sometimes?!
        # for module_name in self.worker_threads:
        #     for command_handle in self.worker_threads[module_name]:
        #         self.worker_threads[module_name][command_handle].wait_shutdown()

    def reload_hook(self):
        # also reload the modules on SIGHUP 
        # remove previous subscription
        self.unsubscribe("control/#")
        self.reload_modules()
        self.on_connect_hook()

    def reload_modules(self):
        print("PiControl :: reload_modules")
        # reset the modules (reloading from the file system)
        # this function is called on __init__ as well!
        # self.modules_internal = {}
        self.modules_external = {}
        # for key in a_dict
        for module_name in self.worker_threads:
            for command_handle in self.worker_threads[module_name]:
                self.worker_threads[module_name][command_handle].set_stale()
        if os.path.isdir(CONFIGPATH):
            modules_list = glob.glob(CONFIGPATH + "*.json")
            for module_path in modules_list:
                if os.path.isfile(module_path):
                    with open(module_path, 'r') as f:
                        try:
                            error, handle, internal, external = self.parse_config(
                                module_path,
                                json.load(f))
                            if handle in self.modules_external:
                                print("picontrol :: handle '" + handle + "' is already in use - not loading module " + module_path)
                            else:
                                if not error:
                                    if "commands" in internal:
                                        for command_handle in internal["commands"]:
                                            # check if we have a worker thread for this already?
                                            flag_add_new = True
                                            if handle in self.worker_threads:
                                                if command_handle in self.worker_threads[handle]:
                                                    flag_add_new = False
                                                    self.worker_threads[handle][command_handle].refresh(config=internal["commands"][command_handle])
                                            else:
                                                self.worker_threads[handle] = {}
                                            if flag_add_new:
                                                self.worker_threads[handle][command_handle] = WorkerThread(
                                                    module_handle=handle,
                                                    command_handle=command_handle,
                                                    config=internal["commands"][command_handle],
                                                    publish_hook=self.publish
                                                )
                                    # only set up the external if the workerThreads could be set up
                                    self.modules_external[handle] = external
                                else:
                                    print("picontrol :: there was an error loading module " + module_path)
                        except Exception as e:
                            print("picontrol :: config " + module_path + " could not be loaded. is it valid JSON? " + str(e))
        else:
            print("picontrol :: CONFIGPATH does not exist: " + CONFIGPATH)
                # for key in a_dict
        for module_name in self.worker_threads:
             for command_handle in self.worker_threads[module_name]:
                self.worker_threads[module_name][command_handle].process_stale_state()

    def parse_config(self, module_path, config):
        # this parses a single configuration file, already loaded as JSON
        error = False
        internal = config
        # this is necessary to create an actual COPY of the data
        # and it NEEDS to be a deepcopy
        external = copy.deepcopy(config)
        handle = "?"
        if "handle" in config:
            handle = config["handle"]
            if "/" in handle or "+" in handle or "#" in handle or "$" in handle:
                error = True
                print("picontrol :: illegal character used in handle " + handle + " in module " + module_path)
        else:
            # handle IS required
            error = True
            print("picontrol :: handle missing in module " + module_path)

        # walk through external and strip command
        if "commands" in external:
            # privacy:
            # if maskcommand is true, we set the command to ""
            for command in external["commands"]:
                if "/" in command or "+" in command or "#" in command or "$" in command:
                    error = True
                    print("picontrol :: illegal character used in command handle " + command + " in module " + module_path)
                if "maskcommand" in external["commands"][command]:
                    if external["commands"][command]["maskcommand"]:
                        if "command" in external["commands"][command]:
                            external["commands"][command]["command"] = ""

        return error, handle, internal, external

    def on_control_msg(self, client, userdata, message):
        # NB: message.topic is valid syntax
        # using it here to prevent fuzzing / DDoS against nonsensical topics
        # extract the control target
        match = self.topic_regex.fullmatch(message.topic)
        if match is not None:
            module_name = match.group(1)
            command_handle = match.group(2)
            if module_name in self.worker_threads:
                if command_handle in self.worker_threads[module_name]:
                    topic, payload = PCPMQTTClient.parse_message(message)
                    self.worker_threads[module_name][command_handle].process_message(topic=topic, payload=payload)
                    # we are currently NOT interested in the contents of the message
                    # every kind of message on this topic triggers the same response
                    # topic, payload = PCPMQTTClient.parse_message(message)

    def on_connect_hook(self):
        # read definition files, and publish.
        # initial definition files are created by installing the picockpit-client package
        # for handle in self.modules_external:
        self.subscribe(
            topic="control/#",
            message_callback_function=self.on_control_msg)
        self.publish(
            topic='$meta',
            message=self.modules_external,
            retain=True,
            overwrite=True)
        for module_name in self.worker_threads:
            for command_handle in self.worker_threads[module_name]:
                self.worker_threads[module_name][command_handle].publish_state()
