import os
import re
import copy
import glob
import json
import errno
import threading
import subprocess

from string import Template
from queue import PriorityQueue
# from picockpit_client.pmqtt import PCPMQTTClient as PCPMQTTClient

CONFIGPATH = '/etc/picockpit-client/apps/com.picockpit/pidoctor/modules/'


class WorkerThread:
    def __init__(self, handle, config, publish_hook):
        self.handle = handle
        self.config = config
        # publish_hook is internal from ComPicockpitPiDoctor, there the topic_app_prefix is set
        self.publish = publish_hook
        self.commands = {}
        # try:
        if "commands" in self.config:
            self.commands = self.setup_commands(self.config["commands"])
        self.wt = threading.Thread(target=self.thread_run, daemon=True)
        # the flag is initially false
        self.run_flag = threading.Event()
        self.q = PriorityQueue()
        self.wt.start()

    def terminate(self):
        self.q.put((1, '$end'))
        self.run_flag.set()

    def setup_commands(self, commands):
        # TODO: recursion on subcommands. 
        # a parser to set up command properties properly, to allow faster responses and less processing time
        for cmd_def in commands:
            if "type" in commands[cmd_def] and "entries" in commands[cmd_def] and "regex" in commands[cmd_def]:
                if commands[cmd_def]["type"] != "array":
                    # force entries to None if type is not an array
                    print("pidoctor :: warning removing entries key from " + cmd_def + " as type is not set to array")
                    commands[cmd_def]["entries"] = None
            else:
                # set the default; this will also happen if the entries key or the regex key are missing!
                commands[cmd_def]["type"] = "single"
                commands[cmd_def]["entries"] = None
            # recurse over the subcommands
            if commands[cmd_def]["entries"]:
                commands[cmd_def]["entries"] = self.setup_commands(commands[cmd_def]["entries"])
            # unset unnecessary values
            commands[cmd_def].pop('name', None)
            commands[cmd_def].pop('description', None)
            # we will also mask errors, to protect spillage!
            if "maskcommand" not in commands[cmd_def]:
                commands[cmd_def]["maskcommand"] = False
            if "ondemand" not in commands[cmd_def]:
                commands[cmd_def]["ondemand"] = False
            if "skiplines" in commands[cmd_def]:
                if not isinstance(commands[cmd_def]["skiplines"], int):
                    print("pidoctor :: warning, for " + cmd_def + " skiplines was not an integer!")
                    commands[cmd_def]["skiplines"] = 0
            else:
                commands[cmd_def]["skiplines"] = 0
            if "timeout" in commands[cmd_def]:
                if not isinstance(commands[cmd_def]["timeout"], int):
                    print("pidoctor :: warning, for " + cmd_def + " timeout was not an integer!")
                    commands[cmd_def]["timeout"] = 5
            else:
                commands[cmd_def]["timeout"] = 5
            if "regex" not in commands[cmd_def]:
                # NB the regex key can also be used for single feedback
                commands[cmd_def]["regex"] = None
            else:
                if "flags" not in commands[cmd_def]["regex"]:
                    commands[cmd_def]["regex"]["flags"] = ""
                if "expression" not in commands[cmd_def]["regex"]:
                    print("pidoctor :: warning, expression key is missing in regex for " + cmd_def)
                    commands[cmd_def]["regex"] = None
            if "substitution" not in commands[cmd_def]:
                commands[cmd_def]["substitution"] = False
            if "literal" not in commands[cmd_def]:
                commands[cmd_def]["literal"] = False
        return commands

    def process_message(self, sub_command):
        if sub_command == "":
            # $ is reserved for picockpit, should never appear as handle!
            self.q.put((3, '$all'))
        elif sub_command == "$all":
            print("pidoctor :: error :: $all is not a valid and allowed handle!")
        elif sub_command == "$end":
            print("pidoctor :: error :: $end is not a valid and allowed handle!")
        else:
            self.q.put((2, sub_command))
        self.run_flag.set()

    def thread_run(self):
        def sanitize(input_string):
            #https://maxharp3r.wordpress.com/2008/05/15/pythons-minidom-xml-and-illegal-unicode-characters/
            RE_XML_ILLEGAL = u'([\u0000-\u0008\u000b-\u000c\u000e-\u001f\ufffe-\uffff])' + \
                             u'|' + \
                             u'([%s-%s][^%s-%s])|([^%s-%s][%s-%s])|([%s-%s]$)|(^[%s-%s])' % \
                             (chr(0xd800), chr(0xdbff), chr(0xdc00), chr(0xdfff),
                              chr(0xd800), chr(0xdbff), chr(0xdc00), chr(0xdfff),
                              chr(0xd800), chr(0xdbff), chr(0xdc00), chr(0xdfff))
            return re.sub(RE_XML_ILLEGAL, "", input_string)

        def run_substitution(sub_dict, content):
            s = Template(content).safe_substitute(sub_dict)
            return s

        def run_cmd(handle, cmd, cmd_def, sub_dict={}):
            # sub_dict is the substitution dictionary taken from the match
            # only required when recursing
            return_code = 0
            return_error = ""
            return_result = ""
            if "command" not in cmd_def:
                print("pidoctor :: error :: command to actually execute is missing in " + handle + " / " + cmd + " definition")
                return_code = errno.ENOSYS
                return_error = "command definition is missing: " + handle + "/" + cmd
            else:
                command = cmd_def["command"]
                if isinstance(command, list):
                    if sub_dict and cmd_def["substitution"]:
                        command = [run_substitution(sub_dict, item) for item in command]
                    try:
                        res = subprocess.run(
                            command,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            timeout=cmd_def["timeout"])
                    except FileNotFoundError:
                        return_code = errno.ENOENT
                        return_error = "Command not found on your system " + " ".join(command)
                    except subprocess.TimeoutExpired:
                        return_code = errno.ETIME
                        return_error = "Command timeout expired (" + str(cmd_def["timeout"]) + " sec)"
                    except Exception as e:
                        if hasattr(e, 'errno'):
                            return_code = e.errno
                        else:
                            return_code = 1
                        if hasattr(e, 'message'):
                            return_error = str(e.message)
                        else:
                            return_error = str(e)
                    else:
                        return_code = res.returncode
                        return_error = sanitize(res.stderr.decode('UTF-8').rstrip("\r\n"))
                        return_result = sanitize(res.stdout.decode('UTF-8').rstrip("\r\n"))
                elif isinstance(command, str):
                    if sub_dict and cmd_def["substitution"]:
                        command = run_substitution(sub_dict, command)
                    if cmd_def["literal"]:
                        # this needs to be parsed with substitution for subcommands
                        # it returns a tuple of returncode, error, output
                        return_result = command
                    else:
                        try:
                            with open(command, 'r') as f:
                                return_result = sanitize(f.read().rstrip("\r\n"))
                        except Exception as e:
                            if hasattr(e, 'errno'):
                                return_code = e.errno
                            else:
                                return_code = 1
                            if hasattr(e, 'message'):
                                return_error = str(e.message)
                            else:
                                return_error = str(e)
                else:
                    print("pidoctor :: error :: command to execute in " + handle + " / " + cmd + " is neither an array nor a string")
                    return_code = errno.EINVAL
                    return_error = "command to execute in " + handle + " / " + cmd + " is neither an array nor a string"
            # postprocess skiplines
            if cmd_def["skiplines"] > 0:
                splits = return_result.split(sep="\n", maxsplit=cmd_def["skiplines"])
                if len(splits) == (cmd_def["skiplines"] + 1):
                    return_result = splits[cmd_def["skiplines"]]
                else:
                    # return an empty string if there are not enough newlines
                    return_result = ""
            # postprocess masking for error messages
            if cmd_def["maskcommand"] and return_error != "":
                return_error = "(error occured, and masked)"
            # postprocess regex
            if cmd_def["regex"] is not None:
                flags = 0
                if "A" in cmd_def["regex"]["flags"]:
                    flags = flags | re.ASCII
                if "S" in cmd_def["regex"]["flags"]:
                    flags = flags | re.DOTALL
                if "I" in cmd_def["regex"]["flags"]:
                    flags = flags | re.IGNORECASE
                if "L" in cmd_def["regex"]["flags"]:
                    flags = flags | re.LOCALE
                if "M" in cmd_def["regex"]["flags"]:
                    flags = flags | re.MULTILINE
                if "X" in cmd_def["regex"]["flags"]:
                    flags = flags | re.VERBOSE
                p = re.compile(cmd_def["regex"]["expression"], flags)
                re_matches = p.finditer(return_result)
                # post-process depending on type of command
                if cmd_def["type"] == "array":
                    return_result = []
                    for match in re_matches:
                        sub_msg = {}
                        sub_dict = match.groupdict()
                        sub_dict['__all'] = match.group(0)
                        for sub_cmd in cmd_def["entries"]:
                            rc_inner, re_inner, rs_inner = run_cmd(
                                handle=handle,
                                cmd=sub_cmd,
                                cmd_def=cmd_def["entries"][sub_cmd],
                                sub_dict=sub_dict)
                            sub_msg[sub_cmd] = {
                                "returncode": rc_inner,
                                "error": re_inner,
                                "output": rs_inner
                            }
                        return_result.append(sub_msg)
                else:
                    # for now we only assume array and single as valid types
                    return_result = ""
                    for match in re_matches:
                        # add the entire match and a newline
                        return_result += match.group(0) + "\n"

            return (return_code, return_error, return_result)

        ld = threading.local()
        ld.run = True
        while ld.run:
            self.run_flag.wait()
            # if run_flag is set, one or more messages have arrived.
            (prio, cmd) = self.q.get()
            if self.q.empty():
                self.run_flag.clear()
            msg = {
                "data": {
                }
            }
            if cmd == '$end':
                # exit
                ld.run = False
            elif cmd == '$all':
                # run all (except on demand tests)
                for cmd in self.commands:
                    if not self.commands[cmd]["ondemand"]:
                        return_code, return_error, return_result = run_cmd(
                            handle=self.handle,
                            cmd=cmd,
                            cmd_def=self.commands[cmd])
                        msg["data"][cmd] = {
                            "returncode": return_code,
                            "error": return_error,
                            "output": return_result
                        }
                # Note, that this publishes on self.handle because these are all commands
                self.publish(
                    topic="$state/" + self.handle,
                    message=msg,
                    overwrite=True)
            else:
                # check if this subtest is in the definition?
                if cmd in self.commands:
                    return_code, return_error, return_result = run_cmd(
                        handle=self.handle,
                        cmd=cmd,
                        cmd_def=self.commands[cmd])
                    msg["data"][cmd] = {
                        "returncode": return_code,
                        "error": return_error,
                        "output": return_result
                    }
                    # Note, that this publishes on self.handle + "/" + cmd because this is an individual command
                    self.publish(
                        topic="$state/" + self.handle + "/" + cmd,
                        message=msg,
                        overwrite=True)
            # go into a blocking state if the queue is empty.


class ComPicockpitPiDoctor:
    def __init__(self, global_run_flag, hooks):
        self.global_run_flag = global_run_flag
        self.hooks = hooks
        self.topic_app_prefix = "com.picockpit/pidoctor/"
        self.topic_regex = re.compile('^pi/[^/]+/com.picockpit/pidoctor/control/([^/]+)/?(.*)')
        self.worker_threads = {}
        # prepare the modules once
        self.reload_modules()

    def subscribe(self, topic, message_callback_function):
        self.hooks["subscribe"](
            topic=self.topic_app_prefix + topic,
            message_callback_function=message_callback_function)

    def unsubscribe(self, topic):
        self.hooks["unsubscribe"](
            topic=self.topic_app_prefix + topic)

    def publish(self, topic, message, retain=False, overwrite=False):
        self.hooks["publish"](
            topic_app_prefix=self.topic_app_prefix,
            topic_proper=topic,
            message=message,
            retain=retain,
            overwrite=overwrite)

    def sigterm_hook(self):
        for module_name in self.worker_threads:
            self.worker_threads[module_name].terminate()

    def timer_hook(self, timer_spacing):
        pass

    def reload_hook(self):
        # also reload the modules on SIGHUP 
        # remove previous subscription
        self.unsubscribe("control/#")
        self.reload_modules()
        self.on_connect_hook()

    def reload_modules(self):
        # reset the modules (reloading from the file system)
        # this function is called on __init__ as well!
        # self.modules_internal = {}
        self.modules_external = {}
        # for key in a_dict
        for wt in self.worker_threads:
            self.worker_threads[wt].terminate()
        self.worker_threads = {}
        if os.path.isdir(CONFIGPATH):
            modules_list = glob.glob(CONFIGPATH + "*.json")
            for module_path in modules_list:
                if os.path.isfile(module_path):
                    with open(module_path, 'r') as f:
                        try:
                            error, handle, internal, external = self.parse_config(
                                module_path,
                                json.load(f))
                            if handle in self.modules_external:
                                print("pidoctor :: handle '" + handle + "' is already in use - not loading module " + module_path)
                            else:
                                if not error:
                                    # set up worker threads, one for each module
                                    self.worker_threads[handle] = WorkerThread(
                                        handle=handle,
                                        config=internal,
                                        publish_hook=self.publish)
                                    # only set up the external if the workerThread could be set up
                                    self.modules_external[handle] = external
                                else:
                                    print("pidoctor :: there was an error loading module " + module_path)
                        except Exception as e:
                            print("pidoctor :: config " + module_path + " could not be loaded. is it valid JSON? " + str(e))
        else:
            print("pidoctor :: CONFIGPATH does not exist: " + CONFIGPATH)

    def parse_config(self, module_path, config):
        # this parses a single configuration file, already loaded as JSON
        error = False
        internal = config
        # this is necessary to create an actual COPY of the data
        # and it NEEDS to be a deepcopy
        external = copy.deepcopy(config)
        handle = "?"
        if "handle" in config:
            handle = config["handle"]
            if "/" in handle or "+" in handle or "#" in handle or "$" in handle:
                error = True
                print("pidoctor :: illegal character used in handle " + handle + " in module " + module_path)
        else:
            # handle IS required
            error = True
            print("pidoctor :: handle missing in module " + module_path)

        # walk through external and strip command
        if "commands" in external:
            # privacy:
            # if maskcommand is true, we set the command to ""
            for command in external["commands"]:
                if "/" in command or "+" in command or "#" in command or "$" in command:
                    error = True
                    print("pidoctor :: illegal character used in command handle " + command + " in module " + module_path)
                if "maskcommand" in external["commands"][command]:
                    if external["commands"][command]["maskcommand"]:
                        if "command" in external["commands"][command]:
                            external["commands"][command]["command"] = ""

        return error, handle, internal, external

    def on_control_msg(self, client, userdata, message):
        # NB: message.topic is valid syntax
        # using it here to prevent fuzzing / DDoS against nonsensical topics
        # extract the control target
        match = self.topic_regex.fullmatch(message.topic)
        if match is not None:
            module_name = match.group(1)
            sub_command = match.group(2)
            if module_name in self.worker_threads:
                # we are currently NOT interested in the contents of the message
                # every kind of message on this topic triggers the same response
                # topic, payload = PCPMQTTClient.parse_message(message)
                self.worker_threads[module_name].process_message(sub_command)

    def on_connect_hook(self):
        # read definition files, and publish.
        # initial definition files are created by installing the picockpit-client package
        # for handle in self.modules_external:
        self.subscribe(
            topic="control/#",
            message_callback_function=self.on_control_msg)
        self.publish(
            topic='$meta',
            message=self.modules_external,
            retain=True,
            overwrite=True)
