import re
import time
import psutil
import datetime
import threading
import subprocess


class ComPicockpitPiStats:
    def __init__(self, global_run_flag, hooks):
        self.global_run_flag = global_run_flag
        self.hooks = hooks
        self.topic_app_prefix = "com.picockpit/pistats/"
        # self.publish = hooks['publish']
        # MULTILINE: ^ and $ match at beginning and end of individual lines
        # note: the whitespace match before link is ON PURPOSE
        self.wlan_regex = re.compile(r"Link Quality.+dBm")
        self.skipped_iwconfig_errors = 0
        self.count_iwconfig_warnings = 0
        self.max_iwconfig_warnings = 5
        # initial setup call
        psutil.cpu_percent(interval=None, percpu=False)
        self.st = threading.Thread(target=self.stats_thread, daemon=True)
        self.run_flag = threading.Event()
        # run_flag controls the execution of the thread;
        # it is set in the on_connect_hook!
        # self.run_flag.set()
        self.st.start()

    def sigterm_hook(self):
        # allow the internal thread which might be blocked to join.
        self.run_flag.set()

    def subscribe(self, topic, message_callback_function):
        self.hooks["subscribe"](
            topic=self.topic_app_prefix + topic,
            message_callback_function=message_callback_function)

    def unsubscribe(self, topic):
        self.hooks["unsubscribe"](
            topic=self.topic_app_prefix + topic)

    def publish(self, topic, message, retain=False, overwrite=True):
        self.hooks["publish"](
            topic_app_prefix=self.topic_app_prefix,
            topic_proper=topic,
            message=message,
            retain=retain,
            overwrite=overwrite)

    def stats_thread(self):
        ld = threading.local()
        while self.global_run_flag.is_set():
            # block until run_flag is true
            # (we have just reconnected or have received a keep alive)
            # this run_flag will also be unblocked on the sigterm_hook being called
            # allow the tread to terminate by falling out at the bottom
            self.run_flag.wait()
            # time.time() returns number of seconds since the epoch
            # using int to cast to an integer without fractions of a second
            ld.ts_last_keep_alive = int(time.time())
            ld.ts_now = int(time.time())
            # three minutes = 180 seconds
            while (ld.ts_now - ld.ts_last_keep_alive < 180) and self.global_run_flag.is_set():
                # the flag will be set to true by the initial setup
                # and by periodic pings from the picockpit frontend
                # whenever it is set to true, we reset the last_keep_alive
                # and clear the flag (set it to false)
                # this way, if no further keep alives arrive, the system
                # will drop into blocking / waiting mode at some point
                if self.run_flag.is_set():
                    ld.ts_last_keep_alive = int(time.time())
                    self.run_flag.clear()
                # sys.stdout.write("✪")
                # sys.stdout.flush()
                self.publish(
                    topic='$state',
                    message=self.get_stats(),
                    retain=False,
                    overwrite=True)
                time.sleep(1)
                ld.ts_now = int(time.time())

    # returns a Dict with the values in it, which can be published directly
    def get_stats(self):
        # CPU info
        # run this first, to "untaint" processing in our function
        stats = {}
        try:
            cpu_usage = psutil.cpu_percent(interval=None, percpu=False)
            stats['cpu_load'] = cpu_usage
        except Exception as e:
            print("picockpit-client :: pistats :: error getting CPU load")
            print(str(e))
        try:
            mem = psutil.virtual_memory()
            stats["ram_total_bytes"] = mem.total
            stats["ram_available_bytes"] = mem.available
        except Exception as e:
            print("picockpit-client :: pistats :: error getting virtual memory")
            print(str(e))
        # get WLAN link quality from iwconfig
        try:
            res = subprocess.run(
                ["iwconfig", "wlan0"],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                timeout=1)
        except Exception as e:
            stats["wlan_signal_info"] = "error running iwconfig"
            # the below could become very verbose very quickly (this is run once per sec)
            # % = modulo
            if self.skipped_iwconfig_errors % 600 == 0 and self.count_iwconfig_warnings < self.max_iwconfig_warnings:
                print("picockpit-client :: pistats :: error running iwconfig (" + str(self.skipped_iwconfig_errors) + ") similar messages")
                print(str(e))
                self.count_iwconfig_warnings += 1
            if self.count_iwconfig_warnings == self.max_iwconfig_warnings:
                print("picockpit-client :: pistats :: maximum number of warnings for iwconfig reached for this session. Will stop logging errors to avoid annoying you and filling up your log.")
                self.count_iwconfig_warnings += 1
            self.skipped_iwconfig_errors += 1
        else:
            wlan_text = res.stdout.decode('UTF-8')
            w_match = self.wlan_regex.search(wlan_text)
            if w_match:
                stats["wlan_signal_info"] = w_match.group(0)
            elif "Access Point: Not-Associated" in wlan_text:
                stats["wlan_signal_info"] = "Not Associated to Access Point"
            else:
                stats["wlan_signal_info"] = ""
        # get temperature
        # "/opt/vc/bin/vcgencmd","measure_temp"
        #exists = os.path.isfile(PATH_VCGENCMD)
        #stats["soc_temperature"] = ""
        #if exists:
        #    res = subprocess.run(
        #        [PATH_VCGENCMD, "measure_temp"],
        #        stdout=subprocess.PIPE,
        #        stderr=subprocess.PIPE)
        #    # rstrip to remove a trailing newline
        #    res_dec = res.stdout.decode('UTF-8').rstrip()
        #    if "temp" in res_dec:
        #        res_dec = str.replace(res_dec, "temp=", "")
        #        res_dec = str.replace(res_dec, "'C", "")
        #        stats["soc_temperature"] = res_dec
        #else:
        #    # fallback to psutil.sensors_temperatures
        try:
            temp = psutil.sensors_temperatures()  # in Celsius
            # The fix below is necessary as depending on the installation, it will be underscore OR dash
            if 'cpu-thermal' in temp:
                stats["soc_temperature"] = temp['cpu-thermal'][0].current
            elif 'cpu_thermal' in temp:
                stats["soc_temperature"] = temp['cpu_thermal'][0].current
            else:
                stats["soc_temperature"] = '?'
        except Exception as e:
            print("picockpit-client :: pistats :: error getting SoC temperature")
            print(str(e))
        # main partition
        try:
            disk_usage = psutil.disk_usage('/')
            stats["root_partition_total"] = disk_usage.total
            stats["root_partition_used"] = disk_usage.used
        except Exception as e:
            print("picockpit-client :: pistats :: error getting disk usage")
            print(str(e))
        # stats["root_partition_free"] = disk_usage.free
        # stats["root_partition_free"] = disk_usage.free
        try:
            disk_io = psutil.disk_io_counters(perdisk=False, nowrap=True)
            stats['disk_io_bytes_read'] = disk_io.read_bytes
            stats['disk_io_bytes_written'] = disk_io.write_bytes
        except Exception as e:
            print("picockpit-client :: pistats :: error getting disk_io")
            print(str(e))        

        try:
            net_io = psutil.net_io_counters(pernic=False, nowrap=True)
            stats['net_io_bytes_sent'] = net_io.bytes_sent
            stats['net_io_bytes_received'] = net_io.bytes_recv
        except Exception as e:
            print("picockpit-client :: pistats :: error getting net_io")
            print(str(e))        

        try:
            now = datetime.datetime.now()
            boot_time = datetime.datetime.fromtimestamp(
                psutil.boot_time())  # time in seconds since the epoch
            delta = now - boot_time
            seconds = delta.seconds
            hours, seconds = divmod(delta.seconds, 3600)
            minutes, seconds = divmod(seconds, 60)
            up = {
                "days": delta.days,
                "hours": hours,
                "minutes": minutes,
                "seconds": seconds
            }
            stats['uptime'] = up
        except Exception as e:
            print("picockpit-client :: pistats :: error getting boot_time")
            print(str(e))        

        return stats

    def timer_hook(self, timer_spacing):
        pass

    def on_control_msg(self, client, userdata, message):
        # this resets the thread's countdown.
        # not interested in the message contents
        self.run_flag.set()

    def reload_hook(self):
        # avoid double subscription (possibly unnecessary here)
        self.unsubscribe(topic="control")
        # this possibly should ensure that the client is connected ..
        # or better: ensure upstream that no errors are generated by this
        self.on_connect_hook()

    def on_connect_hook(self):
        self.subscribe(
            topic="control",
            message_callback_function=self.on_control_msg)
        meta = {
            "uptime": {
                "$description": "Raspberry Pi uptime",
                "minutes": {
                    "$unit": "minute(s)"},
                "days": {
                    "$unit": "day(s)"},
                "seconds": {
                    "$unit": "second(s)"},
                "hours": {
                    "$unit": "hour(s)"}
            },
            "soc_temperature": {
                "$description": "SoC temperature",
                "$unit": "°C"
            },
            "ram_total_bytes": {
                "$description": "RAM total (without VideoCore)",
                "$unit": "bytes"
            },
            "ram_available_bytes": {
                "$description": "RAM available",
                "$unit": "bytes"
            },
            "net_io_bytes_received": {
                "$description": "Network total received data",
                "$unit": "bytes"
            },
            "net_io_bytes_sent": {
                "$description": "Network total sent data",
                "$unit": "bytes"
            },
            "root_partition_used": {
                "$description": "Root partition ('/') total used",
                "$unit": "bytes"
            },
            "root_partition_total": {
                "$description": "Root partition ('/') total available",
                "$unit": "bytes"
            },
            "disk_io_bytes_written": {
                "$description": "Disk I/O total written",
                "$unit": "bytes"
            },
            "disk_io_bytes_read": {
                "$description": "Disk I/O total read",
                "$unit": "bytes"
            },
            "cpu_load": {
                "$description": "CPU load (all cores)",
                "$unit": "%"
            },
            "wlan_signal_info": {
                "$description": "WLAN signal",
                "$unit": ""
            }
        }
        self.publish(
            topic='$meta',
            message=meta,
            retain=True,
            overwrite=True)
        # important: run initial info for three minutes,
        # to avoid missing a run message at startup / reconnect
        self.run_flag.set()
