import com_picockpit.core.core as ComPicockpitCore
import com_picockpit.gpio.gpio as ComPicockpitGPIO
import com_picockpit.picontrol.picontrol as ComPicockpitPiControl
import com_picockpit.pidoctor.pidoctor as ComPicockpitPiDoctor
import com_picockpit.pistats.pistats as ComPicockpitPiStats
import com_picockpit.bme688airmeter.bme688airmeter as ComPicockpitBME688Airmeter
import com_picockpit.bme688digitalnose.bme688digitalnose as ComPicockpitBME688DigitalNose
import com_picockpit.co2sensor.co2sensor as ComPicockpitCO2Sensor
# import com_picockpit.sensors as ComPicockpitSensors
import com_picockpit.resource_manager as ResourceManager


class PluginController:
    # mqtt is a reference to pcp.mqtt_client -> which allows access to MQTT functions directly.
    def __init__(self, global_run_flag, mqtt, upstream_functions={}):
        self.upstream_functions = upstream_functions
        self.plugins = {
        }
        self.mqtt = mqtt
        hooks = {
            "publish": self.mqtt.publish,
            "subscribe": self.mqtt.subscribe,
            "unsubscribe": self.mqtt.unsubscribe
        }
        self.resource_manager = ResourceManager.ResourceManager(hooks, global_run_flag)
        # core has to go first, it initializes Rpi.GPIO for instance
        self.plugins['com.picockpit/core'] = ComPicockpitCore.ComPicockpitCore(
            global_run_flag=global_run_flag,
            resource_manager=self.resource_manager,
            hooks=hooks,
            getpublicip=upstream_functions['getpublicip'],
            terminate=upstream_functions['terminate'])

        # plugins are set up by app here!
        self.plugins['com.picockpit/picontrol'] = ComPicockpitPiControl.ComPicockpitPiControl(
            global_run_flag=global_run_flag,
            hooks=hooks)

        self.plugins['com.picockpit/pidoctor'] = ComPicockpitPiDoctor.ComPicockpitPiDoctor(
            global_run_flag=global_run_flag,
            hooks=hooks)

        self.plugins['com.picockpit/pistats'] = ComPicockpitPiStats.ComPicockpitPiStats(
            global_run_flag=global_run_flag,
            hooks=hooks)

        self.plugins['com.picockpit/bme688airmeter'] = ComPicockpitBME688Airmeter.ComPicockpitBME688Airmeter(
            global_run_flag=global_run_flag,
            resource_manager=self.resource_manager,
            hooks=hooks)

        self.plugins['com.picockpit/bme688digitalnose'] = ComPicockpitBME688DigitalNose.ComPicockpitBME688DigitalNose(
            global_run_flag=global_run_flag,
            resource_manager=self.resource_manager,
            hooks=hooks)

        self.plugins['com.picockpit/co2sensor'] = ComPicockpitCO2Sensor.ComPicockpitCO2Sensor(
            global_run_flag=global_run_flag,
            resource_manager=self.resource_manager,
            hooks=hooks)

        self.plugins['com.picockpit/gpio'] = ComPicockpitGPIO.ComPicockpitGPIO(
            global_run_flag=global_run_flag,
            hooks=hooks)

        # self.resource_manager.connect_plugins(self.plugins)
        self.resource_manager.apps.connect_apps(self.plugins)
        # MQTT will call the hooks once after they are set up if it is connected
        # disconnected respectively.
        self.mqtt.setup_on_connect_callback(self.on_connect_hook)
        self.mqtt.setup_on_disconnect_callback(self.on_disconnect_hook)

    def reload_hook(self):
        # triggered by SIGHUP for a reload of the Apps
        for plugin_name, plugin in self.plugins.items():
            hook = getattr(plugin, "reload_hook", None)
            if callable(hook):
                try:
                    hook()
                except Exception as e:
                    print("☯ plugincontroller :: ✖ " + plugin_name + " error on executing reload_hook" + str(e))

    def timer_hook(self, timer_spacing):
        # timer spacing is currently 1, being ignored
        # this is not precise yet (second precision, there will be a drift!)
        # but good enough for now!
        self.resource_manager.timer_hook(timer_spacing)
        for plugin_name, plugin in self.plugins.items():
            hook = getattr(plugin, "timer_hook", None)
            if callable(hook):
                try:
                    hook(timer_spacing)
                except Exception as e:
                    print("☯ plugincontroller :: ✖ " + plugin_name + " error on executing timer_hook"  + str(e))
            # plugin.timer_hook(timer_spacing)

    def on_connect_hook(self):
        for plugin_name, plugin in self.plugins.items():
            hook = getattr(plugin, "on_connect_hook", None)
            if callable(hook):
                try:
                    hook()
                except Exception as e:
                    print("☯ plugincontroller :: ✖ " + plugin_name + " error on executing on_connect_hook"  + str(e))
            else:
                print("☯ plugincontroller :: ✖ " + plugin_name + " does not have on_connect_hook.")

    def on_disconnect_hook(self):
        # new implementation, which checks if the plugin actually has this hook or not.
        for plugin_name, plugin in self.plugins.items():
            hook = getattr(plugin, "on_disconnect_hook", None)
            if callable(hook):
                try:
                    hook()
                except Exception as e:
                    print("☯ plugincontroller :: ✖ " + plugin_name + " error on executing on_disconnect_hook" + str(e))

    def sigterm_hook(self):
        # new implementation, which checks if the plugin actually has this hook or not.
        # CAVE: The plugins must ensure that they DO NOT BLOCK the shutdown!
        # 2.8.2020: possibly a global locking system is better, where the locks are used as messages.
        for plugin_name, plugin in self.plugins.items():
            hook = getattr(plugin, "sigterm_hook", None)
            if callable(hook):
                try:
                    hook()
                except Exception as e:
                    print("☯ plugincontroller :: ✖ " + plugin_name + " error on executing sigterm_hook" + str(e))

    def announce_shutdown(self):
        # This is a special function which just calls core->shutdown
        # shutdown refers to the shutdown of picockpit-client, not the app.
        print("☯ plugincontroller :: announce_shutdown")
        if 'com.picockpit/core' in self.plugins:
            self.plugins['com.picockpit/core'].announce_shutdown()
        else:
            print("☯ plugincontroller :: ... problem: com.picockpit/core does not exist!")
