import time
import threading
import subprocess
import com_picockpit.constants as CONST
from bme68x import BME68X
import sys

try:
    import RPi.GPIO as GPIO
except RuntimeError:
    print("Error importing RPi.GPIO!  This is probably because you need superuser privileges.  You can achieve this by using 'sudo' to run your script")


class Apps:
    def __init__(self):
        self.app_refs = {}
        self.app_id_table = {}

    def update(self, new_app_list):
        # This function receives an array, and should extract the lookup info for us
        for item in new_app_list:
            self.app_id_table[item['app_id']] = item['path']

    def connect_apps(self, apps):
        self.app_refs = apps

    def get_app_by_id(self, app_id):
        ret = None
        if app_id in self.app_id_table:
            path = self.app_id_table[app_id]
            if path in self.app_refs:
                return self.app_refs[path]
        return ret


class GpioPin:
    def __init__(self, bcm_id, apps):
        self.bcm_id = bcm_id
        self.apps = apps
        self.mode = CONST.GpioCapabilities.UNMANAGED
        self.manager = None

    def get_mode(self):
        return self.mode

    def teardown_pin(self):
        # we can rely on self.manager being the right pin to address when called
        app = self.apps.get_app_by_id(self.manager)
        app.teardown_pin(self.bcm_id)

    def update_pin(self):
        app = self.apps.get_app_by_id(self.manager)
        app.update_pin(self.bcm_id, self.mode)

    def update(self, setup):
        manager_switch = False
        new_manager = None
        if 'manager' in setup:
            new_manager = setup['manager']
            if new_manager != self.manager:
                if new_manager is not None:
                    manager_switch = True
                # notify the old manager for teardown, if it is not None
                if self.manager is not None:
                    self.teardown_pin()
                self.manager = new_manager
        if 'mode' in setup and new_manager is not None:
            new_mode = setup['mode']
            if new_mode != self.mode:
                self.mode = new_mode
                if new_mode == CONST.GpioCapabilities.INPUT:
                    GPIO.setup(self.bcm_id, GPIO.IN)
                    self.update_pin()
                elif new_mode == CONST.GpioCapabilities.OUTPUT:
                    GPIO.setup(self.bcm_id, GPIO.OUT)
                    self.update_pin()
            elif manager_switch:
                # notify new manager, even if pin mode has not changed
                self.update_pin()
        elif 'mode' in setup:
            if setup['mode'] == CONST.GpioCapabilities.UNMANAGED:
                # only clean up if the pin had been in use previously
                if self.manager is not None:
                    if self.mode is not CONST.GpioCapabilities.I2C:
                        GPIO.cleanup(self.bcm_id)
                    # notify the old manager for teardown
                    self.teardown_pin()
                    self.manager = None
                    self.mode = CONST.GpioCapabilities.UNMANAGED


class GpioManager:
    def __init__(self, apps):
        self.apps = apps
        self.i2c_active = False
        print("GPIO initializing ... " + GPIO.VERSION)
        GPIO.setmode(GPIO.BCM)

        # pins will be managed by BCM id
        # BCM0 - BCM45 (on compute module
        # = max. of pins currently on Rpi platform)
        self.gpio_pins = {}
        # this will set up pins 0 to 45
        for bcm_id in range(0, 46):
            self.gpio_pins[bcm_id] = GpioPin(bcm_id=bcm_id, apps=self.apps)

    def cleanup(self):
        # GPIO.cleanup() will set ALL channels to input / no pull, if it is run
        # (and pins had been used)!
        GPIO.cleanup()

    def i2c_available(self):
        # returns true or false whether the I2C pins are available
        # TODO: bounds check -> do these pins actually exist? (is the Pi I2C capable)
        sda_mode = self.gpio_pins[CONST.I2CPins.SDA].get_mode()
        scl_mode = self.gpio_pins[CONST.I2CPins.SDA].get_mode()
        result = False

        if (sda_mode == CONST.GpioCapabilities.UNMANAGED or sda_mode == CONST.GpioCapabilities.I2C ):
            if (scl_mode == CONST.GpioCapabilities.UNMANAGED or scl_mode == CONST.GpioCapabilities.I2C ):
                result = True
        return result

    def update(self, data):
        # we receive the data part of the MQTT message sent by the picockpit-frontend
        for bcm_id in self.gpio_pins.keys():
            if str(bcm_id) in data.keys():
                self.gpio_pins[bcm_id].update(data[str(bcm_id)])
            else:
                self.gpio_pins[bcm_id].update({'mode': CONST.GpioCapabilities.UNMANAGED})

class I2CAddress():
    def __init__(self, address):
        self.address = address
        self.lock = threading.Lock()
        self.owner = None

    def locked(self):
        if self.lock.locked():
            return True
        else:
            return False

    def acquire(self, owner):
        if not self.owner:
            self.owner = owner
            self.lock.acquire()
            return True
        elif self.owner != owner:
            return False
        else:
            return True

    def release(self, owner):
        if not self.owner:
            return True
        elif self.owner != owner:
            return False
        else:
            self.owner = None
            self.lock.release()
            return True

    def get_owner(self):
        return self.owner

class I2CManager:
    def __init__(self, hooks, gpio, global_run_flag):
        self.topic_app_prefix = 'com.picockpit/core/'
        self.publish = hooks['publish']
        self.subscribe = hooks['subscribe']
        self.gpio = gpio
        self.i2c_available = False
        self.i2c_active = False
        self.i2c_addresses = [I2CAddress(i2c_addr) for i2c_addr in range(128)]
        self.global_run_flag = global_run_flag
        self.target_i2c_state_string = '1'
        # self.get_state()
        # this will be driven from core.py
        # self.publish_state()
        self.st = threading.Thread(target=self.switch_thread, daemon=True)
        self.run_flag = threading.Event()
        self.st.start()

    def acquire_i2c_addr(self, i2c_addr: int, app_prefix: str):
        return self.i2c_addresses[i2c_addr].acquire(app_prefix)

    def release_i2c_addr(self, i2c_addr: int, app_prefix: str):
        return self.i2c_addresses[i2c_addr].release(app_prefix)

    def get_owner(self, i2c_addr: int):
        return self.i2c_addresses[i2c_addr].get_owner()

    def is_locked(self, i2c_addr: int):
        return self.i2c_addresses[i2c_addr].locked()

    def switch_thread(self):
        # need to introduce this so we don't freeze everything when I2C needs to be switched
        while self.global_run_flag.is_set():
            self.run_flag.wait()
            # clear the flag immediately
            self.run_flag.clear()
            error_info = ''
            # we need a switch if the flag has been raised.
            try:
                go_ahead = True
                if (self.target_i2c_state_string == '1'):
                    res = subprocess.run(
                        ["lsof", "-wt", "/dev/i2c-1"],
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        timeout=1)
                    if (res.stdout.decode('UTF-8').strip() == ''):
                        pass
                        # reading_state_success = True
                    else:
                        go_ahead = False
                        error_info = 'I2C is busy - cannot switch it off (process ID ' + res.stdout.decode('UTF-8').strip() + ')'
                if go_ahead:
                    res = subprocess.run(
                        ["raspi-config", "nonint", "do_i2c", self.target_i2c_state_string],
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        timeout=1)
                else:
                    pass
            except subprocess.TimeoutExpired: 
                error_info = 'Timeout during execution of I2C switch'
            except Exception as e:
                error_info = 'Error running raspi-config nonint do_i2c to set the I2C state'
            else:
                pass
            self.publish_state(error_info=error_info)

    def publish_state(self, update_only=False, error_info=''):
        i2c_old_available = self.i2c_available
        self.get_state(update_only)
        publish_i2c_state = {
            "i2cAvailable": self.i2c_available,
            "i2cActive": self.i2c_active
        }
        if error_info != '':
            publish_i2c_state['error'] = error_info
        if (i2c_old_available == self.i2c_available and update_only):
            pass
        else:
            self.publish(
                topic_app_prefix=self.topic_app_prefix,
                topic_proper='resources/$state/i2c',
                message=publish_i2c_state,
                retain=True,
                overwrite=True)

    def get_state(self, available_only=False):
        # this gets the current I2C state
        # first look at the availability of the GPIO pins we need
        # note: until we receive an MQTT message about the GPIO pin state, we are in the dark
        self.i2c_available = self.gpio.i2c_available()
        if (not available_only):
            self.i2c_active = False
            # reading_state_success = False
            try:
                res = subprocess.run(
                    ["raspi-config", "nonint", "get_i2c"],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    timeout=1)
            except Exception:
                pass
            else:
                if (res.stdout.decode('UTF-8').strip() == '1'):
                    self.i2c_active = False
                    # reading_state_success = True
                elif(res.stdout.decode('UTF-8').strip() == '0'):
                    self.i2c_active = True
                    # reading_state_success = True
                else:
                    pass

    def set_state(self, desired_state):
        i2c_old_available = self.i2c_available
        i2c_old_active = self.i2c_active
        # update existing state first, including available + active
        self.get_state()
        # note: this might be different than the state we believe we are in 
        # as the user might have switched it manually.
        # compare the desired_state with the new active state
        if (desired_state == self.i2c_active):
            # only check if i2c_old_active is equal to i2c_active
            # if (i2c_old_active != self.i2c_active):
            self.publish_state(True)
        else:
            if (not self.i2c_available):
                # no point in switching if I2C is not available!
                self.publish_state(True)
            else:
                # switch
                self.target_i2c_state_string = '1' # False
                if (desired_state):
                    self.target_i2c_state_string = '0'
                self.run_flag.set()

class ResourceManager:
    def __init__(self, hooks, global_run_flag):
        self.publish = hooks['publish']
        self.subscribe = hooks['subscribe']
        self.apps = Apps()
        self.gpio = GpioManager(self.apps)
        self.i2c = I2CManager(hooks, self.gpio, global_run_flag)

    def __del__(self):
        self.gpio.cleanup()

    def timer_hook(self, timer_spacing):
        pass
