from urllib.parse import urlsplit
import time
import hmac
import json
import hashlib  # https://docs.python.org/3/library/hashlib.html
import requests


class API:
    client_credentials = {}
    api_location = ""
    verbose = False

    def __init__(self, client_credentials, api_location, verbose):
        # client is passed in, so the API can get it for
        if client_credentials and api_location:
            self.client_credentials = client_credentials
            self.api_location = api_location
            self.verbose = verbose
            self.session = requests.Session()
        else:
            raise("client_credentials or api_location empty!")

    def getauthenticationheaders(self):
        the_time = time.time()
        nonce_ts = int(the_time)
        # calculate HMAC
        headers = {
            "nonce_ts": str(nonce_ts),
            "prefix": self.client_credentials["prefix"],
            "bcm_serial": self.client_credentials["serial"],
            "hash_c_two": self.client_credentials["hash_c_two"]
        }
        return headers

    def gethmac(self, nonce_ts, access_method, fullurl, body):
        hco_bytes = bytes.fromhex(self.client_credentials["hash_c_one"])
        # canonical URL
        urlparts = urlsplit(fullurl.lower())
        canurl = urlparts.netloc + urlparts.path + "?" + urlparts.query
        if not body:
            body = "" 
        message = str(nonce_ts) + '!' + access_method + '!' + canurl + '!' + body.decode('utf-8')
        hco_hmac = hmac.new(hco_bytes, message.encode('utf-8'), digestmod=hashlib.sha256).hexdigest()
        return hco_hmac

    def defaultanswer(self):
        return {
            "status_code": 0,
            "body": "",
            "object": {},  # JSON object
            "error": ""
        }

    def parseapianswer(self, answer):
        return_value = self.defaultanswer()
        return_value["status_code"] = answer.status_code
        return_value["body"] = answer.text
        if self.verbose:
            print("**answer.text:**")
            print(answer.text)
            print("****************")
        if 500 <= answer.status_code < 600:
            return_value["error"] = "API Server error (server problem)"
        elif answer.status_code == 401:
            return_value["error"] = "Unauthenticated. Please check your API key.Is it still valid and active?"
        elif answer.status_code == 400:
            return_value["error"] = "Malformed request to server. Have you modified this code? Are you using special characters?"            
        elif answer.status_code == 404:
            return_value["error"] = "Requested resource does not exist."            
        elif answer.status_code == 409: 
            return_value["error"] = "Conflict. The resource you are trying to add already exists, or modification would create problems."            
        elif 402 <= answer.status_code < 500 and answer.status_code != 404:
            return_value["error"] = "Other client-side error."
        try:
            return_value["object"] = json.loads(answer.text)
            if "error" in return_value["object"]:
                return_value["error"] = return_value["object"]["error"]
            if "status_code" in return_value["object"]:
                return_value["status_code"] = return_value["object"]["status_code"]
            # print(type(return_value["object"]))
        except json.JSONDecodeError:
            return_value["error"] = "JSON decoding error (malformed answer from API server)"
            return_value["status_code"] = 900 # our own custom error code
        return return_value

    def apicall(self, suburl="/", method='GET', data=""):
        fullurl = self.api_location + suburl
        headers = self.getauthenticationheaders()
        headers['Content-type'] = 'application/json'
        headers['Accept'] = 'application/json'
        req_p = requests.Request(method, fullurl, json=data, headers=headers)
        req_prepped = req_p.prepare()
        hco_hmac = self.gethmac(
            headers["nonce_ts"],
            method,
            fullurl,
            req_prepped.body)
        req_prepped.headers['hmac'] = hco_hmac
        tries = 10
        retried_flag = False
        for i in range(tries):
            try:
                if self.verbose:
                    print("Trying to talk to the API ... ")
                # this is the only place we directly call into the network
                answer = self.session.send(req_prepped)
            except Exception as e:
                print("*** There was an error while trying to talk to the API: " + e.__class__.__name__)
                print("*** Will retry 10 times, on try # " + str(i + 1))
                retried_flag = True
                if i < tries - 1:  # i is zero indexed
                    # will increasingly wait longer, up to a minute
                    time.sleep(tries * 6)
                    continue
                else:
                    raise  # re-raise the original error
            else:
                if self.verbose or retried_flag:
                    print("... OK Successful communication with the API. ")
                # break terminates the loop (while / for)
                break
        return answer

    def get_mqtt_credentials(self, bcm_serial):
        return self.parseapianswer(
            self.apicall(
                suburl="/pi/" + bcm_serial + "/mqtt/credentials",
                method="GET"))

    def getpublicip(self):
        apianswer = self.apicall(suburl="/publicip", method="GET")
        return self.parseapianswer(apianswer)

    def get_latest_client_version(self):
        return self.parseapianswer(self.apicall(suburl="/latest-client-version", method="GET"))

    def getpi(self, bcm_serial):
        return_value = self.defaultanswer()
        if bcm_serial:
            return_value = self.parseapianswer(self.apicall(suburl="/pi/" + bcm_serial, method="GET"))
        else:
            return_value["error"] = "No BCM serial passed in (client error)"
        return return_value

    # def addtestresult(self, bcm_serial, testresult):
    #     return_value = self.defaultanswer()
    #     if bcm_serial and testresult:
    #         request_data = {
    #             "serial": bcm_serial,
    #             "test_result": testresult #is already a Python / JSON object
    #         }
    #         return_value = self.parseapianswer(self.apicall(suburl="/pidoctortestresult", method="POST", data=request_data))
    #     else:
    #         return_value["error"] = "No BCM serial or testresult passed in (client error)"
    #     return return_value

    def addpi(self, pidata):
        return_value = self.defaultanswer()
        if pidata:
            return_value = self.parseapianswer(self.apicall(suburl="/pi", method="POST", data=pidata))
        else:
            return_value["error"] = "No data provided for pi to add! (client-side error)"
        return return_value
