# import os
import sys
import dbus
import time
import ntplib
import signal
import socket
# only to check if we're the main thread
import threading
import subprocess
import zc.lockfile

from com_picockpit import plugincontroller as ComPicockpitPlugincontroller


class DaemonMode:
    def sync_time(self, timeout=60):
        # determine if we have a network connection, loop for 10 sec if not
        counter = 0
        int_timeout = int(timeout)
        if int_timeout == 0:
            loop_forever = True
            print("picockpit-client :: important info: DNS timeout is set to 0, which will lead to us potentially looping forever here, waiting for DNS to come up.")
        else:
            loop_forever = False
        gotDNS = False
        while not gotDNS:
            counter += 1
            # gotDNS = True
            try:
                socket.getaddrinfo("picockpit.com", 443, proto=socket.IPPROTO_TCP)
            except Exception as e:
                time.sleep(1)
                print("picockpit-client :: cannot get address info for picockpit.com")
                print(str(e))
                if not loop_forever and counter > int_timeout:
                    print("picockpit-client :: DNS startup delay time exceeded, proceeding")
                    gotDNS = True
            else:
                gotDNS = True
        # determine if we need to sync the time (if the time drift is big)
        try:
            net_time_client = ntplib.NTPClient()
            net_time = net_time_client.request('pool.ntp.org')
            local_time = time.time()
            the_time = net_time.tx_time
            print("picockpit-client :: obtained time from pool.ntp.org")
            # print("Picockpit-client :: local_time : " + str(int(local_time)))
            # print("Picockpit-client :: NTP the_time : " + str(int(the_time)))
            if abs((int(local_time))-(int(the_time))) > 30:
                print("picockpit-client :: the time difference is greater than 30 seconds. Updating the clock.")
                try:
                    res = subprocess.run(
                        ["systemctl", "-q", "stop", "systemd-timesyncd"],
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        timeout=1)
                except Exception as e:
                    print("picockpit-client :: error stopping systemd-timesyncd")
                    print(str(e))
                try:
                    res = subprocess.run(
                        ["systemctl", "-q", "start", "systemd-timesyncd"],
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        timeout=1)
                except Exception as e:
                    print("picockpit-client :: error starting systemd-timesyncd")
                    print(str(e))
                print("picockpit-client :: finished syncing the clock.")
                net_time = net_time_client.request('pool.ntp.org')
                local_time = time.time()
                the_time = net_time.tx_time
                # print("picockpit-client :: obtained time from pool.ntp.org")
                # print("Picockpit-client :: new local_time : " + str(int(local_time)))
                # print("Picockpit-client :: new NTP the_time : " + str(int(the_time)))                
        except Exception as e:
            print("picockpit-client :: could NOT obtain time from pool.ntp.org")
            print(str(e))
            the_time = time.time()
            local_time = time.time()

    def sigterm_handler(self, _signo, _stack_frame):
        print("Requested termination with signal " + str(_signo))
        self.terminate()

    def terminate(self):
        # a special for the core plugin
        self.plugins.announce_shutdown()
        # stop plugin threads / while loops
        self.global_run_flag.clear()
        print("Notifying plugins for clean-up activities.")
        # TODO: this would be nicer and cleaner if it was an independent registry of "last wills";
        # Or run in threads or whatever. Otherwise it will fail if one plugin fails
        # this sigterm_hook should set flags of internal blocked threads to run, so they're able to terminate
        self.plugins.sigterm_hook()
        # try:
        #     # sigterm_hook = getattr(self.plugins, "sigterm_hook", None)
        #     # if callable(sigterm_hook):
        #     #     sigterm_hook()
        # except Exception:
        #     print("picockpit-client :: calling sigterm_hook failed somewhere.")
        print("Goodbye. Enjoy your day! Closing lock ...")
        self.lock.close()
        print("calling Dbus")
        system_bus = dbus.SystemBus()
        systemd1 = system_bus.get_object('org.freedesktop.systemd1', '/org/freedesktop/systemd1')
        manager = dbus.Interface(systemd1, 'org.freedesktop.systemd1.Manager')    
        # in s name, in s mode
        manager.StopUnit('picockpit-client.service', 'fail')
        print("calling sys.exit(0)")
        sys.exit(0)

    def reload_handler(self, _signo, _stack_frame):
        print("picockpit-client :: Requested reloading of configuration (SIGHUP)")
        print("picockpit-client :: this will reload individual plugins only, not the whole picockpit-client configuration.")
        # TODO: should we reconnect here to MQTT as well?!
        # TODO: should we reload the config from picockpit-client.conf as well & apply changes?
        self.plugins.reload_hook()

    def run(self):
        while True:
            time.sleep(1)
            # this can probably be implemented in a more predictable way!
            # e.g. actually being in it's own thread, and being called every second
            self.plugins.timer_hook(1)

    def __init__(self, args, pcp, verbose):
        # connect the signals
        self.pcp = pcp

        signal.signal(signal.SIGTERM, self.sigterm_handler)
        signal.signal(signal.SIGINT, self.sigterm_handler)
        # unsure if we really need SIGINT == Ctrl + C
        signal.signal(signal.SIGHUP, self.reload_handler)
        self.args = args
        self.verbose = verbose
        print("Starting picockpit-client in daemon mode.")
        self.sync_time(self.pcp.get_config_key("connection_timeout"))
        # threads in plugins will run while the global_run_flag is true
        self.global_run_flag = threading.Event()
        self.global_run_flag.set()
        # print("I believe the Unix timestamp to be: " + str(int(time.time())))
        # check whether we are the only daemon in town
        try:
            self.lock = zc.lockfile.LockFile('/var/lock/picockpit-client.lock')
        except zc.lockfile.LockError:
            print("Could not obtain lock on /var/lock/picockpit-client.lock - probably another daemon instance is already running.")
            print("Terminating. Sorry :-(")
            sys.exit(1)
        else:
            # initialize the MQTT connection before setting up the plugins
            if self.pcp.mqtt_connect(self.global_run_flag):
                try:
                    try:
                        self.plugins = ComPicockpitPlugincontroller.PluginController(
                            mqtt=self.pcp.mqtt_client,
                            global_run_flag=self.global_run_flag,
                            upstream_functions={
                                "getpublicip": self.pcp.getpublicip,
                                "terminate": self.terminate})
                    except Exception as e:
                        print("daemon_mode :: problem bringing up plugins, terminating. The exception is: " + str(e))
                    else:
                        self.run()
                except Exception as e:
                    print("daemon_mode :: run-time exception, terminating. The exception is: " + str(e))
                finally:
                    self.lock.close()
            else:
                print("MQTT connection not possible - please run 'sudo picockpit-client setupmqtt --verbose' for diagnostics")
                self.lock.close()
                sys.exit(1)
