import os
import re
import dbus
import json
import platform
import threading
import subprocess

from picockpit_client._version import version


class InteractiveMode:
    def handleapierror(self, apianswer):
        print()
        print("The following error was encountered: ")
        print("ERROR: [" + str(apianswer["status_code"]) +"] " + apianswer["error"])
        if self.verbose:
            print("ANSWER BODY:")
            print(apianswer["body"])
        else:
            print("TIP:   Use the --verbose flag to get the server's feedback")

    def addkey(self):
        ak_invalid = True
        if self.args.apikey:
            print("Using API Key passed on command line ...")
            ak_invalid = self.pcp.addapikey(self.args.apikey)
            if ak_invalid:
                print("API Key passed on command line is invalid, please specify another key. Terminating.")
                quit()
        else:
            print("Connecting to picockpit.com. Please enter your API key:")
            while(ak_invalid):
                enteredkey = input('APIKEY: ')
                if not enteredkey:
                    print("Aborting connection as per your request.")
                    quit()
                else:
                    ak_invalid = self.pcp.addapikey(enteredkey)
                    if ak_invalid:
                        print("The key seems to be invalid. Please try again, or press Enter to exit")
                        print("")

    def info(self):
        print("This is PiCockpit Client Version " + version)
        print("********************************************")
        # code copied from pcc.py
        try:
            notread = True
            f = open('/proc/cpuinfo', 'r')
            for line in f:
                if line[0:6] == 'Serial':
                    serial = line[10:26]
                    notread = False
            f.close()
            if notread:
                raise("serial-not-read-error")
            print("Your pi's serial is: " + serial)
        except:
            print("### ERROR: Not able to read the Raspberry Pi serial, which is required for picockpit-client")

        try:
            res = subprocess.run(
                ["vcgencmd", "otp_dump"],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                timeout=1)
        except Exception as e:
            print("### ERROR: not able to read revision from OTP bits: " + str(e))
        else:
            vcgencmd_text = res.stdout.decode('UTF-8')
            self.otp_regex = re.compile(r"30:(.+)$", flags=re.MULTILINE)
            w_match = self.otp_regex.search(vcgencmd_text)
            if w_match:
                revision = w_match.group(1)
                print("Your pi's revision is " + revision)
            else:
                print("### ERROR: not able to match revision from OTP bits.")

        print("Thank you for flying PiCockpit today, we wish you a pleasant stay at your destination.")

    def test(self):
        print("Currently no function happening here. But do definitely have a nice day and come again soon :-)")

    def startservice(self):
        print("Starting picockpit-client daemon service ...")
        system_bus = dbus.SystemBus()
        systemd1 = system_bus.get_object('org.freedesktop.systemd1', '/org/freedesktop/systemd1')
        manager = dbus.Interface(systemd1, 'org.freedesktop.systemd1.Manager')
        manager.EnableUnitFiles(['picockpit-client.service'], False, True)
        manager.Reload()
        job = manager.RestartUnit('picockpit-client.service', 'fail')

    def setupmqtt(self):
        print("Setting up MQTT credentials ...")
        print("(info: this step is necessary for your pi and picockpit.com to be able to communicate bi-directionally)")
        print("The MQTT credentials are tied to your API key. If you delete the API key, they will be deleted as well.")
        apianswer = self.pcp.get_mqtt_credentials()
        if apianswer["error"]:
            self.handleapierror(apianswer)
            print("Quitting. Sorry :-(")
            quit()
        else:
            if self.verbose:
                print("API Server answer: ")
                print(json.dumps(apianswer, indent=4, sort_keys=True))
            print()
            print("OK obtained credentials, testing them.")
            local_run_flag = threading.Event()
            local_run_flag.set()
            testresult = self.pcp.mqtt_connection_test(
                local_run_flag=local_run_flag,
                connectiondata=apianswer["object"])
            print()
            local_run_flag.clear()
            if testresult['success']:
                # save the credentials
                print("Connection test successful. Saving MQTT credentials ... ")
                self.pcp.save_mqtt_credentials(apianswer["object"])
            else:
                print("Connection test was NOT successful:")
                print(testresult['message'])
                print()
                print("MQTT credentials not saved. You can run 'sudo picockpit-client setupmqtt' to try again.")
                print("Quitting. Sorry :-(")
                quit()

    def addpi(self):
            print("Checking whether your Pi is already in your picockpit.com account ...")
            apianswer = self.pcp.getthispi()
            if apianswer["error"] and apianswer["status_code"] != 404:
                self.handleapierror(apianswer)
                print("Quitting. Sorry :-(")
                quit()
            elif apianswer["status_code"] == 404:
                # pi is not in database, suggest to user to add it
                print("This Pi is not in picockpit.com yet.")
                print("")
                proceed = False
                if self.args.yes:
                    proceed = True
                    print("... Adding automatically.")
                else:
                    print("Do you want to add it (required to use picockpit-client properly)? (Y/n)")
                    shall = input(" Y(es)/n(o): ")
                    if not (len(shall) > 0 and shall.strip().lower()[0] != 'y'):
                        proceed = True
                if not proceed:
                        print("Aborting, as requested. Bye!")
                        quit()
                else:
                    print()
                    hostname = platform.node()
                    if self.args.name:
                        print("Using (" + self.args.name + ") as name")
                        name = self.args.name
                    elif self.args.yes:
                        print("Using default: (" + hostname + ") as name")
                        name = hostname
                    else:
                        print("Please add a name for the pi, or simply press Enter to use the default = your Pi's hostname: ")
                        name = input("(" + hostname + "):")
                        if len(name) == 0:
                            name = hostname
                    print()
                    print("Adding your Pi to picockpit.com as '" + name + "' ... ")
                    apianswer = self.pcp.addthispi(name)
                    if apianswer["error"]:
                        self.handleapierror(apianswer)
                        print("Quitting. Sorry :-(")
                        quit()
                    else:
                        print("OK: Your Pi was added to picockpit.com. You should now see it when you log in under My Pi's")
            else:
                print("OK: Your Pi is already saved in your picockpit.com account.")

    def __init__(self, args, pcp, verbose):
        self.args = args
        self.verbose = verbose
        self.pcp = pcp
        print()
        if self.args.command == 'connect':
            if os.geteuid() != 0:
                exit("Please run this script using root privileges (sudo picockpit-client connect).")
            if not self.args.skipkey:
                self.addkey()
            print()
            self.addpi()
            print()
            self.setupmqtt()
            print()
            self.startservice()
            print()
            print("Finished with connect - your Pi is now connected to picockpit.com, if everything was OK")
        elif self.args.command == 'convertcreds':
            self.pcp.convertcredentials_12()
            # automatically terminates
        elif self.args.command == 'info':
            self.info()
        elif self.args.command == 'getprefix':
            print()
            print("The API prefix for the currently saved key is: ")
            print(self.pcp.getprefix())
            print()
            print("Have a nice day!")
        elif self.args.command == 'setupmqtt':
            self.setupmqtt()
            print()
            self.startservice()
            print()            
        elif self.args.command == 'pidoctor':
            print("pidoctor has now been moved to the frontend, please run the tests using the webinterface")
            print("(no data for pidoctor tests is being stored on picockpit.com, the test results are just shown to you.)")
            print(">>> https://www.picockpit.com")
            #     # persist the privacy setting across boots
            #     self.pcp.set_config_key('pidoctor_privacymode', self.args.privacy)
            #     self.pidoctor(self.args.privacy)
        elif self.args.command == 'version':
            self.pcp.get_latest_client_version()
            print("Goodbye, see you again soon! :-)")
            quit()
        elif self.args.command == 'test':
            self.test()
        else:
            print("unknown command + '" + self.args.command + "'")
        self.pcp.get_latest_client_version()
