import os
import re
import json
import pprint
import base64
import hashlib  # https://docs.python.org/3/library/hashlib.html
import binascii
import subprocess
import picockpit_client.secrets
from picockpit_client.api import API
from picockpit_client._version import version
from picockpit_client.pmqtt import PCPMQTTClient
from cryptography.fernet import Fernet
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

DEFAULT_CONFIG_FILE = os.path.dirname(os.path.abspath(__file__)) + "/defaultconfig.json"


class PiCockpitClient:
    serial = ''
    encoded_serial = ''
    config = {}
    client_credentials = {}
    mqtt_credentials = {}
    verbose = False
    VERSION = version

    # salt and token are expected as hex coded bytes
    def symmetric_decode(self, h_salt, h_token):
        if self.serial:
            password = self.serial.encode('utf-8')  # this is implied
            salt = bytes.fromhex(h_salt)
            kdf = PBKDF2HMAC(
                algorithm=hashes.SHA256(),
                length=32,
                salt=salt,
                iterations=100000,
                backend=default_backend())
            key = base64.urlsafe_b64encode(kdf.derive(password))
            f = Fernet(key)
            token = bytes.fromhex(h_token)
            return f.decrypt(token).decode('utf-8')
        else:
            raise("serial-empty-error")

    # provide a salt as hexstring, to chain encoding functions
    def symmetric_encode(self, payload, h_salt=''):
        if self.serial:
            password = self.serial.encode('utf-8')  # this is implied
            if h_salt:
                salt = bytes.fromhex(h_salt)
            else:
                salt = os.urandom(16)
            kdf = PBKDF2HMAC(
                algorithm=hashes.SHA256(),
                length=32,
                salt=salt,
                iterations=100000,
                backend=default_backend())
            key = base64.urlsafe_b64encode(kdf.derive(password))
            f = Fernet(key)
            token = f.encrypt(payload.encode('utf-8'))
            hex_ret = {
                "token": token.hex(),
                "salt": salt.hex()
            }
            return hex_ret
        else:
            raise("serial-empty-error")

    def splash(self):
        print("### picockpit-client " + self.VERSION + " ###")

    def setencodedserial(self):
            # build hash for Pi, using bcmhash_salt
            dk = hashlib.pbkdf2_hmac(
                'sha512',
                self.serial.encode('utf-8'),
                self.config["bcmhash_salt"].encode('utf-8'),
                100000
            )
            self.encoded_serial = binascii.b2a_base64(dk).decode('utf-8').rstrip('\n')        

    def getserial(self, serial):
        if serial:
            print("### overriding BCM serial with user-passed serial: -->" + serial + "<--")
            self.serial = serial
            self.setencodedserial()
        else:
            # read BCM serial, generate secret serial
            try:
                notread = True
                f = open('/proc/cpuinfo', 'r')
                for line in f:
                    if line[0:6] == 'Serial':
                        self.serial = line[10:26]
                        notread = False
                f.close()
                if notread:
                    raise("serial-not-read-error")
            except:
                print("### ERROR: Not able to read the Raspberry Pi serial, which is required for picockpit-client")
                quit()
            else:
                if self.verbose:
                    print("### my BCM serial is: " + self.serial)
                self.setencodedserial()

    def writesettings(self):
        with open(self.config_file_path, "w") as write_file:
            json.dump(self.config, write_file, indent=4, sort_keys=True)

    def updatesettings_coda(self):
        print("Written settings to " + self.config_file_path)
        # refresh credentials, if we just updated the settings file.
        self.setupcredentials()

    def updatesettings(self):
        # this updates the settings file, and reloads the settings if necessary.
        # settings should already be applied to the self.config hash
        # TODO (optionally): validation of self.config hash, to ensure that we don't write a broken configuration
        # TODO (optionally): check that we don't overwrite a file which was changed by the user since the application was started
        try:
            self.writesettings()
        except:
            try: 
                target_directory = os.path.dirname(self.config_file_path)
                print("Creating " + target_directory)
                os.makedirs(target_directory, exist_ok=True)  # recursively create directories
                self.writesettings()
            except:
                print("### ERROR: I was not able to write to the config file: " + self.config_file_path)
                print("Are you running this application as root? (sudo picockpit-client)")
                quit()
            else:
                self.updatesettings_coda()  # success
        else:
            self.updatesettings_coda()  # success

    def printupdatehowto(self):
            print("### If you would like to upgrade, run the following on your command line: ")
            print()
            print("  sudo apt-get update && sudo apt-get --only-upgrade install picockpit-client")
            print()
            print("### to update to a newer version / or check if there is a new one")                

    def get_latest_client_version(self):
        api_answer = {
            "status_code": 901,
            "error": "not authenticated (client credentials missing). run 'sudo picockpit-client connect' first",
            "body": "",
            "object": {}
        }
        if self.client_credentials:
            print()
            print("### Checking for the latest available client version with the API ...")
            api_answer = self.api.get_latest_client_version()
            if api_answer["error"]:
                if self.verbose:
                    print("ERROR: [" + str(api_answer["status_code"]) +"] " + api_answer["error"])
                    print("ANSWER BODY:")
                    print(api_answer["body"])                    
                print("### I can't check the latest client version, as no client credentials are saved.")
                self.printupdatehowto()
            else:
                print("### Your client version is: " + self.VERSION)
                if isinstance(api_answer["object"], str):
                    api_answer["object"] = json.loads(api_answer["object"])
                print("### The latest available version is: " + api_answer["object"]["latest-client-version"])
                print("### The minimum recommended version is: " + api_answer["object"]["minimum-recommended-version"])
                if self.VERSION != api_answer["object"]["latest-client-version"]:
                    self.printupdatehowto()
        else:
            print("### I can't check the latest client version, as no client credentials are saved.")
            self.printupdatehowto()

    def get_mqtt_credentials(self):
        api_answer = {
            "status_code": 901,
            "error": "not authenticated (client credentials missing). run 'sudo picockpit-client connect' first",
            "body": "",
            "object": {}
        }
        if self.client_credentials:
            api_answer = self.api.get_mqtt_credentials(self.serial)

        return api_answer

    def getpublicip(self):
        api_answer = {
            "status_code": 901,
            "error": "not authenticated (client credentials missing). run 'sudo picockpit-client connect' first",
            "body": "",
            "object": {}
        }
        if self.client_credentials:
            api_answer = self.api.getpublicip()

        return api_answer

    def getthispi(self):
        api_answer = {
            "status_code": 901,
            "error": "not authenticated (client credentials missing). run 'sudo picockpit-client connect' first",
            "body": "",
            "object": {}
        }
        if self.client_credentials:
            api_answer = self.api.getpi(self.serial)

        return api_answer

    def addthispi(self, name):
        api_answer = {
            "status_code": 901,
            "error": "not authenticated (client credentials missing). run 'sudo picockpit-client connect' first",
            "body": "",
            "object": {}
        }
        # add this pi by posting it to picockpit.com
        if self.client_credentials:
            if name:
                # read this pi's data - first a minimal setup
                has_lan = True
                has_wlan = True
                try:
                    with open('/sys/class/net/eth0/address') as f:
                        eth0_mac = f.read().rstrip()
                except FileNotFoundError:
                    has_lan = False
                    eth0_mac = "N/A"
                try:
                    with open('/sys/class/net/wlan0/address') as f:
                        wlan0_mac = f.read().rstrip()
                except FileNotFoundError:
                    has_wlan = False
                    wlan0_mac = "N/A"
                with open('/proc/device-tree/model') as f:
                    pi_model = f.read().rstrip('\x00')
                revision = None
                # we need to get the revision from the OTP bit 30
                # this depends on vcgencmd otp_dump being accessible
                try:
                    res = subprocess.run(
                        ["vcgencmd", "otp_dump"],
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        timeout=1)
                except Exception as e:
                    print("... error reading revision from OTP bits ... still proceeding to add the pi.")
                else:
                    vcgencmd_text = res.stdout.decode('UTF-8')
                    self.otp_regex = re.compile(r"30:(.+)$", flags=re.MULTILINE)
                    w_match = self.otp_regex.search(vcgencmd_text)
                    if w_match:
                        revision = w_match.group(1)
                    else:
                        print("... error matching revision from OTP bits ... still proceeding to add the pi.")

                pidata = {
                    "serial": self.serial,
                    "revision": revision,
                    "model": pi_model,
                    "name": name,
                    "lan_mac": eth0_mac,
                    "wlan_mac": wlan0_mac,
                    "has_lan": has_lan,
                    "has_wlan": has_wlan
                }
                api_answer = self.api.addpi(pidata)
            else:
                api_answer["status_code"] = "910"
                api_answer["error"] = "No name provided for pi to be added"
        return api_answer

    def save_mqtt_credentials(self, credentials):
        # the client id is the BCM serial ... no need to save it !
        # self.config["apikeys"][self.encoded_serial]['mqtt_client_id'] = credentials["client_id"]
        # the username is the prefix, no need to save it ! 
        # self.config["apikeys"][self.encoded_serial]['mqtt_username'] = credentials["username"]
        # use cryptography https://cryptography.io/en/latest/fernet/
        hex_ret = self.symmetric_encode(credentials['password'])
        self.config["apikeys"][self.encoded_serial]["mqtt_password"] = hex_ret["token"]
        self.config["apikeys"][self.encoded_serial]["mqtt_salt"] = hex_ret["salt"]
        self.updatesettings()

    def mqtt_connection_test(self, local_run_flag, connectiondata):
        testresult = {
        'success': False,
        'code': -2,
        'message': 'Testing not possible, port and/or host are missing'
        }
        if ('mqtt_server' in self.config) and ('mqtt_port' in self.config):
            pmclient = PCPMQTTClient(
                global_run_flag=local_run_flag,
                connectiondata={
                    "client_id": connectiondata["client_id"],
                    "username": connectiondata["username"],
                    "password": connectiondata["password"],
                    "host": self.config['mqtt_server'],
                    "port": self.config['mqtt_port'],
                    "transport": self.config['mqtt_transport'],
                    "path": self.config['mqtt_path']},
                topicroot="pi/" + self.serial + "/",
                verbose=self.verbose,
                insecure=self.mqtt_selfsigned)
            testresult = pmclient.testconnection()
        return testresult

    def mqtt_connect(self, global_run_flag):
        print("pcc.py :: Connecting to MQTT Service.")
        retval = False
        if not self.mqtt_credentials:
            self.load_mqtt_credentials()
        if not 'mqtt_server' in self.config or not 'mqtt_port' in self.config:
            print("ERROR: mqtt_server or mqtt_port missing in your config file, please check.")
        elif self.mqtt_credentials:
            self.mqtt_client = PCPMQTTClient(
                global_run_flag=global_run_flag,
                connectiondata={
                    "client_id": self.mqtt_credentials["client_id"],
                    "username": self.mqtt_credentials["username"],
                    "password": self.mqtt_credentials["password"],
                    "host": self.config['mqtt_server'],
                    "port": self.config['mqtt_port'],
                    "transport": self.config['mqtt_transport'],
                    "path": self.config['mqtt_path']},
                topicroot="pi/" + self.serial + "/",
                client_string="pi/" + self.serial,
                verbose=self.verbose,
                insecure=self.mqtt_selfsigned)
            # todo: add state here
            self.mqtt_client.will_set(
                topic="$state",
                payload_raw={"onlineState": "offline"},
                qos=2,
                retain=True)
            # setting up callback here.
            # rewrite 25.06.2020 -> no
            # self.mqtt_client.setup_on_connect_callback(on_connect_callback_function)
            info = self.mqtt_client.connect()
            if self.verbose:
                print("MQTT Connection attempt results:")
                self.pp.pprint(info)
            retval = info['success']
        return retval

    def load_mqtt_credentials(self):
        success = False
        self.mqtt_credentials = {}
        if self.client_credentials and self.serial:
            if not 'mqtt_password' in self.config["apikeys"][self.encoded_serial] or not 'mqtt_salt' in self.config["apikeys"][self.encoded_serial]:
                print("ERROR: MQTT Credentials seem not to be present in your settings file")
                print("please set them up using 'sudo picockpit-client setupmqtt'")
            else:
                h_token = self.config["apikeys"][self.encoded_serial]["mqtt_password"]
                h_salt = self.config["apikeys"][self.encoded_serial]["mqtt_salt"]
                password = self.symmetric_decode(h_salt=h_salt, h_token=h_token)
                self.mqtt_credentials = {
                    "client_id": self.serial,
                    "username": self.client_credentials["prefix"],
                    "password": password
                }
                success = True
        return success

    def addapikey(self, fullapikey):
        ak_invalid = True
        fullapikey = fullapikey.strip()
        # check the amount of dots: must be precisely one
        dots = fullapikey.count('.')
        if dots == 1:
            prefix, secretkey = fullapikey.split('.')
            # checking for length of apikey (at least for initial version!.)
            if len(prefix) == 7 and len(secretkey) == 33:
                # now we can do our operations client side
                print("")
                print("Your Pi's serial is: " + self.serial)
                print("This Pi's serial will be automatically used to encrypt the API key on your SD card, to increase security.")
                print("Don't worry: picockpit-client is by default capable of managing the keys for several pis in one settings file.")
                print("You can therefore use the same SD card with another Pi, but will have to run picockpit-client connect again for each additional Pi you add.")
                print("")

                # check if we already have an API entry for this Pi, and prompt the user whether we should overwrite
                if self.encoded_serial in self.config["apikeys"]:
                    proceed_existing = False
                    print("You already have an API key for this Pi. Do you want to replace it? (y/N)")
                    shall = input(" y(es)/N(O): ")
                    if len(shall) > 0 and shall.strip().lower()[0] != 'y':
                        print("Aborting, as requested. Bye!")
                        quit()
                    else:
                        print("OK, replacing the API key.")
                        print()
                m_one = hashlib.sha256()
                m_one.update(fullapikey.encode('utf-8'))
                m_one.update(prefix.encode('utf-8'))
                # here we need to create a re-injection break
                # https://docs.python.org/3/library/hashlib.html
                m_one_digest = m_one.digest()
                m_one_steptwo = hashlib.sha256()
                m_one_steptwo.update(m_one_digest)
                m_one_steptwo.update(self.serial.encode('utf-8'))
                hash_c_one = m_one_steptwo.hexdigest()
                n_two = hashlib.sha256()
                n_two.update(fullapikey.encode('utf-8'))
                n_two.update(fullapikey.encode('utf-8'))
                n_two.update(prefix.encode('utf-8'))
                hash_c_two = n_two.hexdigest()

                self.finalizecredentials(
                    hash_c_one=hash_c_one,
                    hash_c_two=hash_c_two,
                    prefix=prefix)
                ak_invalid = False
        return ak_invalid

    def getprefix(self):
        retval = ''
        if self.client_credentials:
            retval = self.client_credentials['prefix']
        return retval

    def finalizecredentials(self, hash_c_one, hash_c_two, prefix):
                # second step: symmetric encoding
                hr_one = self.symmetric_encode(payload=hash_c_one,h_salt='')
                hr_two = self.symmetric_encode(payload=hash_c_two,h_salt=hr_one['salt'])
                hr_prefix = self.symmetric_encode(payload=prefix,h_salt=hr_one['salt'])

                # now save to the settings file
                # this will also reset the API keys if there had been any.
                self.config["apikeys"][self.encoded_serial] = {
                    "hash_c_one": hr_one['token'],  # hash_c_one,
                    "hash_c_two": hr_two['token'],  # hash_c_two,
                    "prefix": hr_prefix['token'],  # prefix,
                    "mqtt_password": "",
                    "mqtt_salt": "",
                    "cred_salt": hr_one['salt']
                }
                self.updatesettings()

    # wrapper function to update credentials to new v0.12 format
    # CAVE: this will overwrite mqtt settings to empty - as no other clients had them, we will ALLOW this
    def convertcredentials_12(self):
        if not self.client_credentials:
            print("Nothing to convert - you do not seem to have credentials saved for this pi")
            quit()
        elif "cred_salt" in self.config["apikeys"][self.encoded_serial]:
            print("Nothing to be done - your credentials are already in the new format")
            print("Quitting. Have a nice day!")
            quit()
        else:
            self.finalizecredentials(
                hash_c_one=self.client_credentials['hash_c_one'],
                hash_c_two=self.client_credentials['hash_c_two'],
                prefix=self.client_credentials['prefix'])
            print("Converted your credentials - they will be available on the next run of picockpit-client")
            print("This operation only needed to be run once - you will not need to run it again.")
            print("Thank you for upgrading to more secure settings and have a nice day!")
            quit()

    def setupcredentials(self):
        self.client_credentials = {}
        # check if the current Pi is in the loaded settings, if so, set the hash. otherwise unset it
        if self.serial != "" and self.encoded_serial != "":
            if self.encoded_serial in self.config["apikeys"]:
                # adding second step, symmetric decoding
                # compatibility layer
                if "cred_salt" in self.config["apikeys"][self.encoded_serial]:
                    prefix = self.symmetric_decode(
                        h_salt=self.config["apikeys"][self.encoded_serial]["cred_salt"],
                        h_token=self.config["apikeys"][self.encoded_serial]["prefix"])
                    hash_c_one = self.symmetric_decode(
                        h_salt=self.config["apikeys"][self.encoded_serial]["cred_salt"],
                        h_token=self.config["apikeys"][self.encoded_serial]["hash_c_one"])
                    hash_c_two = self.symmetric_decode(
                        h_salt=self.config["apikeys"][self.encoded_serial]["cred_salt"],
                        h_token=self.config["apikeys"][self.encoded_serial]["hash_c_two"])
                else:
                    print("### ")
                    print("### Using legacy compatibility mode to read credentials")
                    print("### We strongly advise you to convert them to the new encrypted setup using")
                    print("### 'sudo picockpit-client convertcreds'")
                    print("### ")
                    prefix = self.config["apikeys"][self.encoded_serial]["prefix"]
                    hash_c_one = self.config["apikeys"][self.encoded_serial]["hash_c_one"]
                    hash_c_two = self.config["apikeys"][self.encoded_serial]["hash_c_two"]
                self.client_credentials = {
                    "serial": self.serial,
                    "prefix": prefix,
                    "hash_c_one": hash_c_one,
                    "hash_c_two": hash_c_two
                }
                self.api = API(self.client_credentials, self.config["api_location"], self.verbose)

    def readsettings(self, config_file_path):
        # read default configuration. it is required in any case.
        if os.path.isfile(DEFAULT_CONFIG_FILE):
            with open(DEFAULT_CONFIG_FILE, 'r') as f:
                self.defaultconfig = json.load(f)
        else:
            print("No default config file found. Terminating. Sorry :-(")
            print("Try reinstalling picockpit-client if you continue to get this error.")
            quit()
        # try to read configuration
        self.config_file_path = config_file_path
        exists = os.path.isfile(config_file_path)

        if exists:
            print("Reading configuration from: " + config_file_path)
            with open(config_file_path, 'r') as f:
                self.config = json.load(f)
        else:
            print("No config file found. Will use built-in default settings.")
            self.config = self.defaultconfig

        if self.config['bcmhash_salt'] == '?':
            bcmhash_salt = picockpit_client.secrets.token_urlsafe(45)
            self.config['pcc_version'] = version
            self.config['bcmhash_salt'] = bcmhash_salt
            # this will write a settings file by default if none exists
            self.updatesettings()

        # update verbosity
        if 'verbose' in self.config:
            if self.config['verbose']:
                self.verbose = True  # force it on, otherwise cmdline
        if 'mqtt_selfsigned' in self.config:
            if self.config['mqtt_selfsigned']:
                self.mqtt_selfsigned = True

    def set_config_key(self, key, value):
        # the key must exist (currently required) in our "default namespace"
        if key in self.defaultconfig:
            flag_update = True
            if key in self.config:
                oldvalue = self.config[key]
                if oldvalue == value:
                    flag_update = False
            if flag_update:
                self.config[key] = value
                self.updatesettings()
        else:
            raise Exception("could not set key " + key + " - it does not exist in the default config")

    # this method used to obtain config information for code outside of this file.
    def get_config_key(self, key):
        if key in self.config:
            return self.config[key]
        else:
            # fall back on default config
            # will fail automatically if key is not in default config either!
            return self.defaultconfig[key]

    def __init__(self, config_file_path, args, daemon_mode=False):
        self.verbose = args.verbose
        self.args = args
        self.daemon_mode = daemon_mode
        self.mqtt_selfsigned = args.selfsigned
        serial = self.args.serial
        #  startup info (only when run interactively?)
        self.splash()
        #  obtain settings, or set up default settings
        self.readsettings(config_file_path)
        #  obtain BCM serial
        # requires bcmhash_salt, therefore readsettings must go first
        self.getserial(serial)
        self.setupcredentials()  # requires read settings, and read serial.
        self.mqtt_client = None
        self.pp = pprint.PrettyPrinter(indent=4)
