import ssl
import json
# import lmdb
import uuid
import time
import queue
import msgpack
import datetime
import threading
import paho.mqtt.client as mqtt


class PCPMQTTClient:

    @staticmethod
    def parse_message(message):
        # we can also verify the message here (integrity, authenticity, etc.)
        topic = message.topic
        if len(message.payload) == 0:
            # empty message
            payload = {}
        else:
            try:
                # new 20.7.2020: msgpack
                payload = msgpack.unpackb(message.payload)
                # payload = json.loads(message.payload)
            except Exception as e:
                print("PCPMQTTClient :: parse_message :: error parsing message " + str(e))
                # prevents fuzzing by wrong JSON
                # possibly we should indicate to upstream function - but no need for now, internal dev.
                payload = {}
        return (topic, payload)

    def on_message(self, client, userdata, message):
        pass

    def unsubscribe(self, topic):
        # todo: this should be implemented in a way where an individual app may unsubscribe, 
        # but we can still stay subscribed for other interested apps.
        topic = self.topicroot + topic
        if self.verbose:
            print("pmqtt.py :: unsubscribe called for computed topic " + topic)
        retval = False
        if topic in self.subscriptions:
            result, mid = self.mqttc.unsubscribe(topic)
            retval = False
            if result == mqtt.MQTT_ERR_SUCCESS:
                retval = True
                if self.verbose:
                    print("pmqtt.py :: subscription was successful!")
            else:
                if self.verbose:
                    print("pmqtt.py :: subscription FAILED!")
        return retval

    def subscribe(self, topic, message_callback_function, qos=1):
        topic = self.topicroot + topic
        if self.verbose:
            print("pmqtt.py :: subscribe called for computed topic " + topic + " with qos " + str(qos))
        self.mqttc.message_callback_add(topic, message_callback_function)
        # https://www.eclipse.org/paho/clients/python/docs/#subscribe-unsubscribe
        # may be called in three different ways as per doc, we only support
        # subscription to one topic per call.
        # result is MQT_ERR_SUCCESS to indicate success
        # mid -> message id für subscribe request, to track in on_subscribe
        # when broker has acknowledged the request, on_subscribe is generated.
        retval = True
        if topic not in self.subscriptions:
            result, mid = self.mqttc.subscribe(topic, qos)
            retval = False
            if result == mqtt.MQTT_ERR_SUCCESS:
                retval = True
                if self.verbose:
                    print("pmqtt.py :: subscription was successful!")
            else:
                if self.verbose:
                    print("pmqtt.py :: subscription FAILED!")
        # else:
        #     # remove the hook!
        #     # cave: what about several functions subscribing to the same path?!
        #     self.mqttc.message_callback_remove()
        return retval

    def setup_on_connect_callback(self, on_connect_callback_function):
        self.on_connect_callback_function = on_connect_callback_function
        self.on_connect_callback_function_active = True
        # run this the first time
        if self.connected_flag:
            on_connect_callback_function()

    def setup_on_disconnect_callback(self, on_disconnect_callback_function):
        self.on_disconnect_callback_function = on_disconnect_callback_function
        self.on_disconnect_callback_function_active = True
        # run this the first time
        if not self.connected_flag:
            on_disconnect_callback_function()

    def on_disconnect(self, client, userdata, rc):
        if self.verbose:
            print("►►►► on_disconnect MQTT")
        if self.on_disconnect_callback_function_active:
            self.on_disconnect_callback_function()

    def on_connect(self, client, userdata, flags, rc):
        if self.verbose:
            print("▲▲▲▲ on_connect MQTT ▲▲▲▲")
        # empty out subscriptions, to start again.
        for topic in self.subscriptions:
            self.mqttc.message_callback_remove(topic)
        self.subscriptions = []

        # This function is also going to be called on reconnect ...,
        # and should therefore trigger re-publishing of our vitality messages
        # (e.g. if disconnect happened due to momentary network outage, etc !)
        self.connection_status = rc
        if rc == 0:
            self.connected_flag = True
            self.connection_error_flag = False
            if self.on_connect_callback_function_active:
                self.on_connect_callback_function()
        else:
            self.connected_flag = False
            self.connection_error_flag = True
            if self.on_disconnect_callback_function_active:
                self.on_disconnect_callback_function()

    # called when a mesage that was to be sent using publish()
    # has completed transmission to the broker
    # for qos=0 -> message has left the client
    # for qos=1 & 2 -> handshakes have been completed.
    # mid variable matches mid variable which publish() returns
    def on_publish(self, client, userdata, mid):
        if self.verbose:
            print("publish callback mid: " + str(mid))

    def prepare_payload(self, payload_raw):
        if self.verbose:
            print("pmqtt.py :: prepare_payload start")
        # payload_raw is a Dict, check that it indeed is
        if isinstance(payload_raw, dict):
            if '$ts' in payload_raw:
                raise Exception('$ts is a reserved value, only used for timestamp!')
            else:
                payload_raw["$ts"] = datetime.datetime.utcnow().timestamp()

            if '$uuid' in payload_raw:
                raise Exception('$uuid is a reserved value, only used for the UUID!')
            else:
                payload_raw["$uuid"] = str(uuid.uuid1())  # https://docs.python.org/3/library/uuid.html
                # TODO: clock_seq can be set and managed for the connection, e.g. incremented on each message
                # TODO: this is for uuid.uuid1

            if '$client' in payload_raw:
                raise Exception('$client is a reserved value, only used for the client!')
            else:
                payload_raw["$client"] = self.client_string

            if self.verbose:
                print("pmqtt.py :: iterating over the items of payload_raw:")
                num = 0
                for key, value in payload_raw.items():
                    num += 1
                    print(str(num) + ' :: ' + str(key) + ' -> ' + str(value))
                print("pmqtt.py :: going through msgpack")
            # 20.07.2020 new: package as msgpack
            packed = msgpack.packb(payload_raw)
            if self.verbose:
                print("pmqtt.py :: packed size (in bytes): " + str(len(packed)))
            return packed
            # return json.dumps(payload_raw)
        else:
            raise Exception('MQTT payload must be a Python dict')

    def on_log(self, client, userdata, level, buf):
        # logging is only turned on if verbose is set in setup
        print(buf)

    def testconnection(self):
        return self.connect(test=True)

    def will_set(self, topic, payload_raw=None, qos=0, retain=False):
        payload = self.prepare_payload(payload_raw)
        self.mqttc.will_set(self.topicroot + topic, payload, qos, retain)

    def connect(self, test=False):
        if test:
            print("Connecting to MQTT", end='')
        self.connection_status = -1  # no connection attempts yet
        self.connection_error_flag = False
        self.mqttc.connect_async(
            host=self.connectiondata['host'],
            port=self.connectiondata['port'],
            keepalive=30,
            bind_address="")
        self.mqttc.loop_start()  # according to Steve better AFTER connect
        continue_flag = False
        keep_connection_flag = not test
        timeout_counter = 0
        while not continue_flag:
            time.sleep(1)
            timeout_counter += 1
            if test:
                print(".", end='')  # end='' will suppress newline
            if self.connected_flag:
                continue_flag = True
            if timeout_counter > 6:
                continue_flag = True
                keep_connection_flag = False
        code = self.connection_status
        success = False
        if code == 0:
            success = True
            message = "Connection successful!"
            self.connected_flag = True
            # publish the first message to activate the pi
            if test:
                self.publish(
                    topic_app_prefix='',
                    topic_proper='$state',
                    message={"onlineState": "offline"},
                    qos=2,
                    retain=True,
                    overwrite=True)
            if not keep_connection_flag:  # necessary to be this, to detect if we are not able to connect!
                self.disconnect()  # clean disconnect from server
        elif code == -1:
            message = "Connection timed out - server not online? Certificate error (try --selfsigned)?"
        elif code == 1:
            message = "Connection refused - incorrect protocol version"
        elif code == 2:
            message = "Connection refused - invalid client identifier"
        elif code == 3:
            message = "Connection refused - server unavailable"
        elif code == 4:
            message = "Connection refused - bad username or password"
        elif code == 5:
            message = "Connection refused - not authorised"
        else:
            message = "Another error " + str(code) + " occured, please check the Paho documentation"
        print('')
        return {"success": success, "code": code, "message": message}

    def publish_thread(self):
        # This thread looks for new messages in the queue and publishes them.
        while self.global_run_flag.is_set():
            # todo: we must unlock this when global_run_flag is set to false
            self.publish_data_present_flag.wait()
            # go through the queue and publish the messages
            while not self.publish_queue.empty():
                if self.verbose:
                    print("pmqtt.py :: fetching new message from publish queue ")
                msg = self.publish_queue.get()
                if self.verbose:
                    print("pmqtt.py :: publishing ... ")
                pubresult = self.mqttc.publish(
                                topic=msg["topic"],
                                payload=msg["payload"],
                                qos=msg["qos"],
                                retain=msg["retain"])
                mid = pubresult.mid
                if self.verbose:
                    print("pmqtt.py :: mid is: " + str(mid))
                # process pubresult
                if pubresult.rc == mqtt.MQTT_ERR_SUCCESS:
                    code = 'MQTT_ERR_SUCCESS'
                    error_message = 'Successfully sent message to broker.'
                    success = True
                elif pubresult.rc == mqtt.MQTT_ERR_NO_CONN:
                    code = 'MQTT_ERR_NO_CONN'
                    error_message = 'No connection to MQTT server, try to reconnect'
                elif pubresult.rc == mqtt.MQTT_ERR_QUEUE_SIZE:
                    code = 'MQTT_ERR_QUEUE_SIZE'
                    error_message = 'Message is neither queued nor sent.'
                if self.verbose:
                    print("pmqtt.py :: publishing result of the above message: ")
                    print({"success": success, "code": code, "message": error_message, "mid": mid})
                mid = pubresult.mid
                if self.verbose:
                    print("pmqtt.py :: mid is: " + str(mid))
                # process pubresult
                if pubresult.rc == mqtt.MQTT_ERR_SUCCESS:
                    code = 'MQTT_ERR_SUCCESS'
                    error_message = 'Successfully sent message to broker.'
                    success = True
                elif pubresult.rc == mqtt.MQTT_ERR_NO_CONN:
                    code = 'MQTT_ERR_NO_CONN'
                    error_message = 'No connection to MQTT server, try to reconnect'
                elif pubresult.rc == mqtt.MQTT_ERR_QUEUE_SIZE:
                    code = 'MQTT_ERR_QUEUE_SIZE'
                    error_message = 'Message is neither queued nor sent.'
                if self.verbose:
                    print("pmqtt.py :: publishing result of the above message: ")
                    print({"success": success, "code": code, "message": error_message, "mid": mid})
            if self.publish_queue.empty():
                self.publish_data_present_flag.clear()

    def publish(self,
                topic_app_prefix,
                topic_proper,
                message,
                qos=1,
                retain=False,
                overwrite=False):
        # overwrite: in case we are offline & messages are queued, overwrite all previous messages for this topic
        success = False
        code = 'PCP_WRONG_INPUTFORMAT'
        error_message = 'Wrong input format of data (e.g. topic cannot be Null)'
        mid = None
        if topic_proper:
            if self.verbose:
                print("pmqtt.py :: publishing to " + self.topicroot + topic_app_prefix + topic_proper)
                print("... input message: ")
                print(message)
                print("")
            # add to the queue instead of directly publishing, and set the flag
            # pubresult = self.mqttc.publish(
            #     topic=self.topicroot + topic_app_prefix + topic_proper,
            #     payload=self.prepare_payload(message),
            #     qos=qos,
            #     retain=retain)
            self.publish_queue.put({
                "topic": self.topicroot + topic_app_prefix + topic_proper,
                "payload": self.prepare_payload(message),
                "qos": qos,
                "retain": retain})
            if self.verbose:
                print("pmqtt.py :: after adding to queue, setting the publish_data_present_flag")
            self.publish_data_present_flag.set()
        # TODO: we have no idea whether the message is going to be published successfully!
        else:
            if self.verbose:
                print("pmqtt.py :: topic_proper is missing, empty or Null, not publishing")
        # TODO: do not rely on these returns -> we have to publish through a queue!
        return {"success": success, "code": code, "message": error_message, "mid": mid}

    def disconnect(self):
        # as per steve http://www.steves-internet-guide.com/client-connections-python-mqtt/
        self.mqttc.loop_stop()
        self.mqttc.disconnect()
        self.connected_flag = False

    def __init__(
            self,
            global_run_flag,
            connectiondata,
            topicroot,
            client_string="!client_not_set!",
            verbose=False,
            insecure=False):
        self.global_run_flag = global_run_flag
        self.subscriptions = []
        self.verbose = verbose
        clean_session = True
        self.mqttc = mqtt.Client(
            client_id=connectiondata['client_id'],
            clean_session=clean_session,
            userdata=None,
            protocol=mqtt.MQTTv311,
            transport=connectiondata['transport'])
        self.connectiondata = connectiondata
        self.connected_flag = False
        # this is used to provide a callback hook when we get disconnected
        # see setup_callback
        self.on_connect_callback_function_active = False
        self.on_disconnect_callback_function_active = False
        self.mqttc.on_connect = self.on_connect
        self.mqttc.on_disconnect = self.on_disconnect
        self.mqttc.on_publish = self.on_publish
        self.mqttc.on_message = self.on_message
        self.topicroot = topicroot
        self.client_string = client_string
        if self.verbose:
            print("### enabling logging for MQTT Client ###")
            self.mqttc.on_log = self.on_log
        self.mqttc.username_pw_set(
            username=connectiondata['username'],
            password=connectiondata['password'])
        if connectiondata['transport'] == "websockets":
            self.mqttc.ws_set_options(path=self.connectiondata['path'])
        if self.verbose:
            print("### connecting with the following parameters ###")
            print("* client_id: " + connectiondata['client_id'])
            print("* clean_session: " + str(clean_session))
            print("* transport: " + connectiondata['transport'])
            print("* path: " + connectiondata['path'])
            print("* username: " + connectiondata['username'])
            print("* password: (not being shown!)")
        if insecure:
            self.mqttc.tls_set(cert_reqs=ssl.CERT_NONE)
            if self.verbose:
                print("### insecure operation mode (--selfsigned) set! ###")
            self.mqttc.tls_insecure_set(True)
        else:
            self.mqttc.tls_set()
        # construct a Thread
        self.pt = threading.Thread(target=self.publish_thread, daemon=True)
        self.publish_queue = queue.Queue()
        self.publish_data_present_flag = threading.Event()
        self.pt.start()
